#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import glob
import os
import types
import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
from . set_style import set_plot_style
from .. pygeostat_parameters import Parameters

@set_plot_style
def histogram_plot_simulation(simulated_data, reference_data, reference_variable=None, reference_weight=None, reference_n_sample=None,
                            simulated_column=None, griddef=None, nreal=None,
                            n_subsample=None, simulated_limits=False, ax=None,
                            figsize=None, xlim=None, title=None, xlabel=None, stat_blk='all',
                            stat_xy=(0.95, 0.05), reference_color=None, simulation_color=None, alpha=None, lw=1,
                            plot_style=None, custom_style=None, output_file=None, out_kws=None, sim_kws=None,
                            **kwargs):
    """
    histogram_plot_simulation emulates the pygeostat histogram_plot program as a means of checking histogram
    reproduction of simulated realizations to the original histogram. The use of python generators
    is a very flexible and easy means of instructing this plotting function as to what to plot.

    The function accepts five types of simulated input passed to the ``simulated_data`` argument:

        #. 1-D array like data (numpy or pandas) containing 1 or more realizations of simulated
           data.
        #. 2-D array like data (numpy or pandas) with each column being a realization and each row
           being an observation.
        #. List containing location(s) of realization file(s).
        #. String containing the location of a folder containing realization files. All files in
           the folder are read in this case.Can contain
        #. String with a wild card search (e.g., './data/realizations/*.out')
        #. Python generator object that yields a 1-D numpy array.

    The function accepts two types of reference input passed to the ``reference_data`` argument:

        #. Array like data containing the reference variable
        #. String containing the location of the reference data file (e.g., './data/data.out')

    This function uses pygeostat for plotting and numpy to calculate statistics.

    The only parameters required are ``reference_data`` and ``simulated_data``. If files are to be read or a 1-D
    array is passed, the parameters ``griddef`` and ``nreal`` are required. ``simulated_column`` is required
    for reading files as well. It is assumed that an equal number of realizations are within each
    file if multiple file locations are passed. Sub-sampling of datafiles can be completed by
    passing the parameter ``n_subsample``. If a file location is passed to ``reference_data``, the parameters
    ``reference_variable`` and ``reference_n_sample`` are required. All other arguments are optional or determined
    automatically if left at their default values. If ``xlabel`` is left to its default value of
    ``None``, the column information will be used to label the axes if present. Three keyword
    dictionaries can be defined. (1) ``sim_kws`` will be passed to pygeostat histogram_plot used for
    plotting realizations (2) ``out_kws`` will be passed to the pygeostat exportfig function and
    (3) ``**kwargs`` will be passed to the pygeostat histogram_plot used to plot the reference data.


    Two statistics block sets are available: ``'minimal'`` and the default ``'all'``. The
    statistics block can be customized to a user defined list and order. Available statistics are
    as follows:

    >>> ['nreal', 'realavg', 'realavgstd', 'realstd', 'realstdstd', 'ndat', 'refavg', 'refstd']

    Please review the documentation of the :func:`gs.set_plot_style() <pygeostat.plotting.set_plot_style>` and
    :func:`gs.export_image() <pygeostat.plotting.export_image>` functions for details on their
    parameters so that their use in this function can be understood.

    Parameters:
        simulated_data: Input simulation data
        reference_data: Input reference data

    Keyword Arguments:
        reference_variable (int, str): Required if sub-sampling reference data. The column containing the data
            to be sub-sampled
        reference_weight: 1D dataframe, series, or numpy array of declustering weights for the data. Can also
            be a string of the column in the reference_data if reference_data is a string, or a bool if reference_data.weights
            is a string
        reference_n_sample (int): Required if sub-sampling reference data. The number of data within the
            reference data file to sample from
        griddef (GridDef): A pygeostat class GridDef created using :class:`gs.GridDef
            <pygeostat.data.grid_definition.GridDef>`
        simulated_column (int): column number in the simulated data file
        nreal (int): Required if sub-sampling simulation data. The total number of realizations
            that are being plotted. If a HDF5 file is passed, this parameter can be used to limit
            the amount of realizations plotted (i.e., the first ``nreal`` realizations)
        n_subsample (int): Required if sub-sampling is used. The number of sub-samples to draw.
        ax (mpl.axis): Matplotlib axis to plot the figure
        figsize (tuple): Figure size (width, height)
        xlim (float tuple): Minimum and maximum limits of data along the x axis
        title (str): Title for the plot
        xlabel (str): X-axis label
        stat_blk (str or list): Indicate what preset statistics block to write or a specific list
        stat_xy (str or float tuple): X, Y coordinates of the annotated statistics in figure
            space. The default coordinates specify the bottom right corner of the text block
        reference_color (str): Colour of original histogram
        simulation_color (str): Colour of simulation histograms
        alpha (float): Transparency for realization variograms (0 = Transparent, 1 = Opaque)
        lw (float): Line width in points. The width provided in this parameter is used for the
            reference variogram, half of the value is used for the realization variograms.
        plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
            :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
            to turn it off
        custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected
        output_file (str): Output figure file name and location
        out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
            :func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`
        sim_kws: Optional dictionary of permissible keyword arguments to pass to
            :func:`gs.histogram_plot() <pygeostat.plotting.histogram_plot.histogram_plot>` for plotting realization
            histograms and by extension, matplotlib's plot function if the keyword passed is not
            used by :func:`gs.histogram_plot() <pygeostat.plotting.histogram_plot.histogram_plot>`
        **kwargs: Optional dictionary of permissible keyword arguments to pass to
            :func:`gs.histogram_plot() <pygeostat.plotting.histogram_plot.histogram_plot>` for plotting the reference
            histogram and by extension, matplotlib's plot function if the keyword passed is not
            used by :func:`gs.histogram_plot() <pygeostat.plotting.histogram_plot.histogram_plot>`

    Returns:
        ax (ax): matplotlib Axes object with the histogram reproduction plot

    **Examples:**

    .. plot:: 
    
        import pygeostat as gs
        import pandas as pd

        # Global setting using Parameters
        gs.Parameters['data.griddef'] = gs.GridDef([10,1,0.5, 10,1,0.5, 2,1,0.5])
        gs.Parameters['data.nreal'] = nreal = 100
        size = gs.Parameters['data.griddef'].count();

        reference_data = pd.DataFrame({'value':np.random.normal(0, 1, size = size)})

        # Create example simulated data
        simulated_data =pd.DataFrame(columns=['value'])
        for i, offset in enumerate(np.random.rand(nreal)*0.04 - 0.02):
            simulated_data= simulated_data.append(pd.DataFrame({'value':np.random.normal(offset, 1, size = size)}))

        gs.histogram_plot_simulation(simulated_data, reference_data, reference_variable='value',
                                title='Histogram Reproduction', grid=True)
    """
    # -----------------------------------------------------------------------
    #  Sanity checks, file type determination, and try loading fortran
    # -----------------------------------------------------------------------
    import pygeostat as gs
    from . utils import format_plot, _set_stat_fontsize
    # Figure out what type of input simulated_data is
    subsamp = False
    ndim = False
    generator = False
    folder = False
    wildcard = False
    array = False
    filelist = False
    if isinstance(simulated_data, types.GeneratorType):
        nreal = 0
        generator = True
        iterator = simulated_data
    elif isinstance(simulated_data, str) and ('*' in simulated_data):
        wildcard = True
    elif isinstance(simulated_data, str) and os.path.isdir(simulated_data):
        folder = True
    elif isinstance(simulated_data, list):
        filelist = True
    elif any([isinstance(simulated_data, pd.DataFrame), isinstance(simulated_data, np.ndarray), isinstance(simulated_data, pd.Series)]):
        array = True
        if subsamp:
            raise ValueError("Sub-sampling won't work if the data is already in memory")
    else:
        raise ValueError("The passed `simulated_data` is not a valid input format")
    if n_subsample is not None and isinstance(n_subsample, (int, float)):
        subsamp = True
    # Make sure the required parameters are passed
    if griddef is None:
        griddef = Parameters['data.griddef']
        if griddef is None:
            raise ValueError("A gs.GridDef must be passed when reading from files")

    if nreal is None:
            nreal = Parameters['data.nreal']
            if nreal is None:
                raise ValueError("The number of realizations to be read must be specified when"
                                 " reading from files")

    if any([folder, wildcard, filelist]):
        
        if simulated_column is None:
            raise ValueError("The column in the files that contains the simulation data must be"
                             " specified")
    elif array:
        if (len(simulated_data.shape) == 1) or (simulated_data.shape[1] == 1):
            if not isinstance(griddef, gs.GridDef):
                raise ValueError("If a 1-D array is passed, a gs.GridDef must be passed to"
                                 " `griddef`")
            if nreal is None:
                raise ValueError("The number of realizations must be passed if dealing with a"
                                 " 1-D array")
    # Figure out what type of input reference_data is
    if isinstance(reference_data, str) and (n_subsample is not False):
        refsubsamp = True
    else:
        refsubsamp = False
    # Try to load the subsample function if it required
    if subsamp or refsubsamp:
        if not isinstance(griddef, gs.GridDef):
            raise ValueError("A gs.GridDef is required for subsampling.")
        try:
            from pygeostat.fortran.subsample import subsample
        except:
            raise ImportError("The fortran subroutine subsample could not be loaded, please ensure"
                              " it has been compiled correctly.")

    # -----------------------------------------------------------------------
    #  Handle data input
    # -----------------------------------------------------------------------
    # Set-up variables
    realavg, realavgstd, realstd, realstdstd, refavg, refstd = ([] for i in range(6))
    # Handle pd and np input
    if array:
        if isinstance(simulated_data, pd.DataFrame):
            simulated_data = simulated_data.values
        if (len(simulated_data.shape) == 2) and (simulated_data.shape[1] > 1):
            ndim = 2
        else:
            ndim = 1
        # Handle 1-D arrays
        if ndim == 1:
            ncell = griddef.count()
        # Handle 2-D arrays
        if ndim == 2:
            nreal = simulated_data.shape[1]
    # Handle folder and wildcard searches
    if folder:
        if simulated_data[-1] != '/':
            simulated_data = simulated_data + '/'
        simulated_data = simulated_data + '*'
        files = []
        for filepath in glob.glob(simulated_data):
            files.append(filepath)
    if wildcard:
        files = []
        for filepath in glob.glob(simulated_data):
            files.append(filepath)
    if filelist:
        files = simulated_data

    if any([folder, wildcard, filelist]):
        ncell = griddef.count()
        ndim = 1
        # Check and make sure the number of files and the nreal value passed makes sense
        if nreal % len(files) != 0:
            raise ValueError(" The number of realizations passed is not divisible by the number"
                             " of files passed/found. Please make sure there are the same number"
                             " of realizations in each file and that the sum of them match the"
                             " nreal argument passed.")
        # Read the data
        simulated_data = []
        for file in files:
            data = gs.DataFile(file).data.iloc[:, simulated_column - 1]
            if simulated_limits:
                if isinstance(simulated_limits, (int, float)):
                    data = data.loc[data > simulated_limits]
                elif len(simulated_limits) == 2:
                    data = data.loc[data.between(simulated_limits[0], simulated_limits[1])]
                else:
                    raise ValueError('simulated_limits must be a value or tuple')
            simulated_data.extend(data)
        simulated_data = np.array(simulated_data)
        if simulated_limits:
            ncell = len(data)
    # Handle sub-sampling if required
    if subsamp:
        ncell = griddef.count()
        if array:
            if ndim == 1:
                simulated_data = np.reshape(simulated_data, (nreal, griddef.count())).T
                ndim = 2
            newarr = np.zeros((n_subsample, nreal))
            for ireal in range(nreal):
                ridx = np.random.permutation(ncell)[:n_subsample]
                newarr[:, ireal] = simulated_data[ridx, ireal]
            simulated_data = newarr
        else:
            # Sub-sample all of the files and combine into a single numpy array
            files = simulated_data
            file_nreal = int(nreal / len(files))
            simulated_data = []
            for fl in files:
                dump = subsample(fl, simulated_column, ncell, n_subsample, file_nreal, gs.rseed())
                dump = np.transpose(dump)
                simulated_data.extend(dump)
            simulated_data = np.array(simulated_data)
            simulated_data = np.transpose(simulated_data)
            nreal = simulated_data.shape[1]

    # Create a generator for the realizations
    if generator is False:
        def _itersimulated_data():
            for i in range(0, nreal):
                if subsamp or ndim == 2:
                    real = simulated_data[:, i]
                elif ndim == 1:
                    real = simulated_data[(i * ncell):(((i + 1) * ncell) - 1)]
                yield real
        iterator = _itersimulated_data()

    # -----------------------------------------------------------------------
    #  Plot Figure
    # -----------------------------------------------------------------------
    # Set figure style parameters

    # Handle dictionary defaults
    if sim_kws is None:
        sim_kws = dict()
    if out_kws is None:
        out_kws = dict()
    # Set-up figure
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    if simulation_color is None:
        simulation_color = Parameters['plotting.histogram_plot_simulation.simclr']
    if alpha is None:
        alpha = Parameters['plotting.histogram_plot_simulation.alpha']
    # Plot realization histograms
    for real in iterator:
        # Append intermediate realization dist statistics to variables
        if stat_blk:
            realavg.append(np.nanmean(real))
            realstd.append(np.nanstd(real))
        if generator:
            nreal += 1
        gs.histogram_plot(real, ax=ax, icdf=True, lw=(lw / 2), stat_blk=False, color=simulation_color, alpha=alpha, plot_style=False, **sim_kws)
    # Calculate more realization dist statistics if required
    if stat_blk:
        realavgstd = np.std(realavg)
        realavg = np.mean(realavg)
        realstdstd = np.std(realstd)
        realstd = np.mean(realstd)
    # Sub-sample original distribution if needed

    if isinstance(reference_data, str):
        if subsamp:
            reference_data = subsample(reference_data, reference_variable, reference_n_sample, n_subsample, 1, gs.rseed())
            reference_data = reference_data[:, 0]
        else:
            reference_data = gs.DataFile(reference_data)
            if isinstance(reference_variable, str):
                reference_variable = reference_data.gscol(reference_variable) - 1
            if isinstance(reference_weight, str):
                reference_weight = reference_data.gscol(reference_weight) - 1
                reference_weight = reference_data.data.values[:, reference_weight]
            elif isinstance(reference_weight, bool):
                if reference_weight:
                    if isinstance(reference_data.weights, str):
                        reference_weight = reference_data[reference_data.weights]
                    else:
                        raise ValueError('reference_weight=True is only valid if reference_data.weights is a string!')
            reference_data = reference_data.data.values[:, reference_variable]
    elif isinstance(reference_data, gs.DataFile):
        if isinstance(reference_weight, str):
            reference_weight = reference_data[reference_weight].values
        elif isinstance(reference_weight, bool):
            if reference_weight:
                if isinstance(reference_data.weights, str):
                    reference_weight = reference_data[reference_data.weights].values
                else:
                    raise ValueError('reference_weight=True is only valid if reference_data.weights is a string!')
        if isinstance(reference_variable, str):
            reference_data = reference_data[reference_variable].values
        elif isinstance(reference_data.variables, str):
            reference_data = reference_data[reference_data.variables].values
        elif len(list(reference_data.columns)) == 1:
            reference_data = reference_data.data.values
        else:
            raise ValueError('could not coerce reference_data into a 1D array!')

    # Plot the reference histogram
    if reference_color is None:
        reference_color = Parameters['plotting.histogram_plot_simulation.refclr']
    if not isinstance(reference_data, bool) and (reference_data is not None):
        gs.histogram_plot(reference_data, weights=reference_weight, ax=ax, icdf=True, stat_blk=False, color=reference_color,
                   lw=lw, plot_style=False, **kwargs)
    # Calculate reference dist statistics if required
    if stat_blk:
        if not isinstance(reference_data, bool) and (reference_data is not None):
            refavg = gs.weighted_mean(reference_data, reference_weight)
            refstd = np.sqrt(gs.weighted_variance(reference_data, reference_weight))
            ndat = len(reference_data)
        else:
            refavg = np.nan
            refstd = np.nan
            ndat = np.nan
    # Configure plot
    if xlabel is None:
        xlabel = gs.get_label(reference_data)
        if xlabel is None:
            xlabel = gs.get_label(simulated_data)
    # axis_xy and grid are applied by format_plot based on the current Parameters setting
    # no kwarg in this function for now since it's already loaded
    ax = format_plot(ax, xlabel, 'Cumulative Frequency', title, xlim=xlim)
    # Ensure that we have top spline, in case it was removed above
    ax.spines['top'].set_visible(True)
    # Plot statistics block
    if stat_blk:
        statlist = {'nreal': (r'$n_{real} = %0.0f$' % nreal),
                    'realavg': (r'$m_{real} = %0.3f$' % realavg),
                    'realavgstd': (r'$\sigma_{m_{real}} = %0.3f$' % realavgstd),
                    'realstd': (r'$\sigma_{real} = %0.3f$' % realstd),
                    'realstdstd': (r'$\sigma_{\sigma_{real}} = %0.3f$' % realstdstd),
                    'ndat': ('$n_{ref} = %0.0f$' % ndat),
                    'refavg': (r'$m_{ref} = %0.3f$' % refavg),
                    'refstd': (r'$\sigma_{ref} = %0.3f$' % refstd)}
        statsets = {'all': ['nreal', 'realavg', 'realavgstd', 'realstd', 'realstdstd', 'ndat',
                            'refavg', 'refstd'],
                    'minimal': ['nreal']}
        if subsamp:
            statlist['n_subsample'] = '$n_{subsample} = %0.0f$' % n_subsample
            statsets['all'].append('n_subsample')
        txtstats, stat_xy, ha, va = gs.get_statblk(stat_blk, statsets, statlist, stat_xy)
        stat_fontsize = _set_stat_fontsize(None)
        ax.text(stat_xy[0], stat_xy[1], txtstats, va=va, ha=ha, fontsize=stat_fontsize,
                transform=ax.transAxes)
    # Export figure
    if output_file or ('pdfpages' in out_kws):
        gs.export_image(output_file, **out_kws)

    return ax



"""
The :mod:`pygslib.datautils` module includes data utility functions.
"""
from __future__ import absolute_import, division, print_function
from .desurvey import *
from .labels import *
from .utils import *



'''
The :mod:`pygslib.transformations` module is for calculating GSLIB rotations and transformations.
'''
from __future__ import absolute_import, division, print_function

from .rotations import *
from .transformations import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
h5_io.py: Contains input/output functions for using HDF5 files within pygeostat
'''
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import os
import numpy as np
import pandas as pd
import textwrap

from ..utility.logging import printerr


def _fixh5path(h5path):
    """Fix an h5path for use in h5py file objects"""
    if h5path in [None, '', ' ', '/']:
        h5path = '/'
    if h5path[0] != '/':
        h5path = '/' + h5path
    if h5path[-1] != '/':
        h5path = h5path + '/'

    return h5path


def h5_combine_data(flname, h5paths, datasets=None):
    """
    Combine data into one DataFrame from multiple paths in a HDF5 file.

    Parameters:
        flname (str): Path of the HDF5 you wish to read from
        h5paths (list): A list of h5paths to combine. Forward slash (/) delimited path through the
            group hierarchy you wish to place the dataset(s) specified by the argument ``datasets``
            into. The dataset name cannot be passed using this argument, it is interpreted as a
            group name. A value of ``None`` places the dataset into the root directory of the
            HDF5 file.
        datasets (list of lists): If only a specific set of datasets from each path are desired then
            pass a list of lists of equal length as the h5paths list. An empty list within the list
            will cause all datasets in the corresponding path to be readin.

    Returns:
        DataFrame

    Example:

    >>> flname = 'drilldata.h5'
    ... h5paths = ['/Orig_data/series4870/', 'NS/Declus/series4870/']
    ... datasets = [['LOCATIONX', 'LOCATIONY', 'LOCATIONZ'], []]
    ... data = gs.h5_combine_data(flname, h5paths, datasets=datasets)
    """
    import h5py
    # Handle some default parameters
    if isinstance(h5paths, list):
        if len(h5paths) > 1:
            for path in h5paths:
                if path in [None, True, '', ' ', '/']:
                    path = '/'
        else:
            printerr("Multiple paths were not passed. If you are not combining data across paths"
                     "use `gs.read_h5()`", errtype='error')
    else:
        printerr("Multiple paths were not passed. If you are not combining data across paths use"
                 " `gs.read_h5()`", errtype='error')
    # Sanity checks
    if not h5py.is_hdf5(flname):
        raise IOError("The passed file path is not a valid HDF5 file.")

    for idx, path in enumerate(h5paths):
        if len(datasets[idx]) > 0:
            if idx == 0:
                data = read_h5(flname=flname, h5path=path,
                               datasets=datasets[idx])
            else:
                data2 = read_h5(flname=flname, h5path=path,
                                datasets=datasets[idx])
                data = pd.concat([data, data2], axis=1)
        else:
            if idx == 0:
                data = read_h5(flname=flname, h5path=path)
            else:
                data2 = read_h5(flname=flname, h5path=path)
                data = pd.concat([data, data2], axis=1)

    return data


def write_h5(data, flname, h5path=None, datasets=None, dtype=None, gridstr=None,
               trim_variable=None, var_min=-998.0):
    """Write data to an HDF5 file using the python package H5PY. The file is appended to and in
    the case that a dataset already exists, it is overwritten.

    Parameters:
        data: A 1-D np.array/pd.Series or a ``pd.DataFrame`` containing different variables as
            columns
        flname (str): Path of the HDF5 you wish to write to or create
        h5path (str): Forward slash (/) delimited path through the group hierarchy you wish to
            place the dataset(s) specified by the argument ``datasets`` into. The dataset name
            cannot be passed using this argument, it is interpreted as a group name. A value of
            ``None`` places the dataset into the root directory of the HDF5 file.
        datasets (str or list): Name of the dataset(s) to write out. If a ``pd.DataFrame`` is
            passed, the values passed by the argument ``datasets`` must match the DataFrame's
            columns.
        dtype (str): The data type to write. Currently, only the following values are permitted:
            ``['int32', 'float32', 'float64']``. If a ``pd.DataFrame`` is passed and this argument
            is left to it's default value of ``None``, the DataFrame's dtypes must be of the types
            listed above.
        gridstr (str): Grid definition string that is saved to the HDF5 file as an attribute
            of the group defined by the parameter ``h5path``.
        trim_variable (str): Variable to use for trimming the data. An index will be written to the
            h5file and will be used to rebuild dataset while only nontrimmed data will be written
            out
        var_min (float): minimum trimming limit usedif trim_variable is passed

    Examples:
        Write a single pd.Series or np.array to an HDF5 file:

        >>> gs.write_h5(array, 'file.h5', h5path='Modeled/Var1', datasets='Realization_0001')

        Write a whole ``pd.DataFrame`` in group (folder) 'OriginalData' that contains a dataset
        for every column in the ``pd.DataFrame``:

        >>> gs.write_h5('file.h5', DataFrame, h5path='OriginalData')
    """
    import h5py
    # Sanity checks
    if not isinstance(data, (np.ndarray, pd.Series, pd.DataFrame)):
        raise TypeError("The data passed must be a np.ndarray, pd.Series, or pd.DataFrame")
    if isinstance(data, (np.ndarray, pd.Series)) and ((isinstance(datasets, list)) and
                                                      (len(datasets) > 1)):
        raise TypeError("The passed data is 1-D yet more than one variable is specified")

    # Sort out some defaults
    if h5path in [None, '', ' ', '/']:
        h5path = '/'

    if datasets is None:
        if isinstance(data, pd.DataFrame):
            datasets = list(data.columns)
        elif isinstance(data, pd.Series):
            datasets = [data.name]
        else:
            datasets = ['data']
    else:
        if not isinstance(datasets, list):
            datasets = [datasets]

    # More sanity checks
    if isinstance(data, pd.DataFrame) and datasets is not None:
        errcols = set(datasets) - set(data.columns)
        if len(errcols) > 0:
            raise KeyError("The following datasets cannot be found in the pd.DataFrame: %s" %
                           errcols)

    # Open up the h5 file
    with h5py.File(flname, 'w') as h5file:
        # Create the groups
        if h5path == '/':
            group = h5file
        else:
            group = h5file.require_group(h5path)

        trim = False
        if trim_variable:
            if isinstance(data, pd.DataFrame):
                array = data[trim_variable]
                trim = True
            elif isinstance(data, pd.Series):
                array = data
                trim = True
            else:
                printerr('Data needs to be either pd.DataFrame, or pd.Series to trim'
                         'Data will be written out untrimmed', errtype='warning')

        if trim:
            h5_index = array.index[array > var_min].values.astype('int32')
            # Make sure the last cell is always included
            last_index = len(array) - 1
            if h5_index.max() < last_index:
                h5_index = np.append(h5_index, [last_index])
            array = np.atleast_2d(h5_index).T
            group.require_dataset(name='h5_index', data=array, shape=array.shape,
                                  dtype='int32')
        # Iterate over the data if required and write it to the HDF5 data
        for dataset in datasets:
            if isinstance(data, pd.DataFrame):
                array = data[[dataset]]
                dtype = str(array.dtypes[0])
            else:
                array = data
            if trim:
                array = array.loc[h5_index]
            # Determine default data type if required
            if dtype is None:
                dtype = str(array.dtype)
            if dtype == 'int64' or dtype == 'int32' or dtype == 'float32' or dtype == 'float64':
                if dtype == 'int64':
                    printerr("The data type 'int64' is not supported by the Fortran module"
                             " hdf5_io. The data will be written as 'int32'.",
                             errtype='warning')
                    array = array.astype('int32')
                array = np.atleast_2d(array).T
                group.require_dataset(name=dataset, data=array, shape=array.shape, dtype=dtype)
            else:
                raise NotImplementedError("The data type `%s` is not supported by the Fortran"
                                          " module hdf5_io. Supported data types: ['int32',"
                                          "'float32', 'float64']" % dtype)

        # Write the griddef if required
        if isinstance(gridstr, bytes):
            h5file[h5path].attrs['griddef'] = '%s' % gridstr


def read_h5(flname, h5path=None, datasets=None, fill_value=-999):
    """
    Return a 1-D array from an HDF5 file or build a ``pd.DataFrame()`` from a list of datasets in a
    single group.

    The argument ``h5path`` must be a path to a group. If 1 or more specific variables are desired
    to be loaded, pass a list to ``datasets`` to specify which to read.

    Parameters:
        flname (str): Path of the HDF5 you wish to write to or create
        h5path (str): Forward slash (/) delimited path through the group hierarchy you wish to
            read the dataset(s) specified by the argument ``datasets`` from. The dataset name
            cannot be passed using this argument, it is interpreted as a group name only. A
            value of ``None`` places the dataset into the root directory of the HDF5 file. A value
            of ``False`` loads a blank pd.DataFrame().
        datasets (str or list): Name of the dataset(s) to read from the group specified by
            ``h5path``. Does nothing if ``h5path`` points to a dataset.
        fill_value (float or np.NaN): value to fill in grid with if trimmed data was written out.
            default is -999

    Returns:
        data (pd.DataFrame): DataFrame containing one or more columns, each containing a single
        1-D array of a variable.
    """
    import h5py
    # Handle some default parameters
    trim = False
    if h5path in [None, True, '', ' ', '/']:
        h5path = '/'
    elif h5path is False:
        return pd.DataFrame()
    else:
        h5path = '/' + h5path
    if isinstance(datasets, str):
        datasets = [datasets]

    # Sanity checks
    if not h5py.is_hdf5(flname):
        raise IOError("The passed file path is not a valid HDF5 file.")

    # Open the file and read the data
    with h5py.File(flname, 'r') as h5store:
        # Sort out some defaults and make sure all the datasets required are present
        if isinstance(h5store[h5path], h5py.Group):
            group = True
            h5dsets = [x for x in list(h5store[h5path].keys()) if
                       isinstance(h5store['%s/%s' % (h5path, x)], h5py.Dataset)]
            if 'h5_index' in h5dsets:
                trim = True
                h5dsets.remove('h5_index')
            if datasets is None:
                datasets = h5dsets
            else:
                error = set(datasets) - set(h5dsets)
                if len(error) > 0:
                    raise KeyError("The following datasets cannot be found in h5path specified: %s"
                                   % error)
        elif isinstance(h5store[h5path], h5py.Dataset):
            group = False
            if h5path.endswith('/'):
                h5path = h5path[:-1]
            datasets = [h5path.rsplit('/', 1)[1]]
            if 'h5_index' in datasets:
                trim = True
                datasets.remove('h5_index')
        else:
            raise ValueError("The `h5path` does not point to a dataset or a group.")

        # Read the data
        data = []
        if trim:
            if group:
                path = h5path + '/h5_index'
            else:
                path = '/h5_index'
            # python
            h5_index = h5store[path][:].flatten()

        for dataset in datasets:
            if group:
                path = h5path + '/' + dataset
            else:
                path = h5path
            # python
            data.append(h5store[path][:].flatten())
        if len(data) > 0:
            data = np.column_stack(data)
            if trim:
                data = pd.DataFrame(data, columns=datasets, index=h5_index)
                data = data.reindex(index=range(h5_index.max() + 1), fill_value=fill_value)
            else:
                data = pd.DataFrame(data, columns=datasets)
        else:
            printerr("No data was found within the root directory. A value of `None` has been"
                     " returned", errtype='warning')
            data = None
    return data


def ish5dataset(h5fl, dataset, h5path=None):
    """
    Check to see if a dataset exits within an HDF5 file

    The argument ``h5path`` must be a path to a group and cannot contain the dataset name.
    Can only check for one dataset at a time.

    Parameters:
        flname (str): Path of the HDF5 you wish to check
        h5path (str): Forward slash (/) delimited path through the group hierarchy you wish to
            check for the specified dataset. The dataset name cannot be passed using this
            argument, it is interpreted as a group name only. A value of ``None`` places the
            dataset into the root directory of the HDF5 file.
        dataset (str): Name of the dataset to check for in the group specified by ``h5path``.

    Returns:
        exists (bool): Indicator if the specified dataset exists
    """
    import h5py
    # Handle some default parameters
    if h5path in [None, True, '', ' ', '/']:
        h5path = '/'
    else:
        h5path = '/' + h5path + '/'
    h5path = h5path + dataset
    # Sanity checks
    if os.path.isfile(h5fl):
        if not h5py.is_hdf5(h5fl):
            raise IOError("The passed file exists but s not a valid HDF5 file.")
        # Check for the dataset
        with h5py.File(h5fl, 'r') as h5store:
            if dataset in [x for x in h5store.keys()]:
                exists = isinstance(h5store[h5path], h5py.Dataset)
            else:
                exists = False
    else:
        exists = False

    return exists


class H5Store:
    """
    A simple class within pygeostat to manage and use HDF5 files.

    :ivar str flname:  Path to a HDF5 file to create or use
    :ivar ``h5py.File`` h5data: h5py File object
    :ivar dict paths: Dictionary containing all of the groups found in the HDF5 file that
        contain datasets

    Parameters:
        flname (str): Path to a HDF5 file to create or use

    Usage:

        Write a np.array or pd.Series to the HDF5 file:

        >>> H5Store['Group1/Group2/Var1'] = np.array()

        Write all the columns in a ``pd.DataFrame`` to the HDF5 file:

        >>> H5Store['Group1/Group2'] = pd.DataFrame()

        Retrieve a single 1-D array:

        >>> array = H5Store['Group1/Group2/Var1']

        Retrieve a single 1-D array within the root directory of the HDF5 file:

        >>> array = H5Store['Var1']

        Retrieve the first value from the array:

        >>> value = H5Store['Var1', 0]

        Retrieve a slice of values from the array:

        >>> values = H5Store['Var1', 10:15]
    """
    import h5py

    def __init__(self, flname, replace=False):

        self.flname = flname
        if self.h5py.is_hdf5(flname):
            if replace:
                os.remove(flname)
                self.h5data = None
                self.paths = None
            else:
                self._loadfile()
        else:
            self.h5data = None
            self.paths = None

    def __str__(self):
        """
        Print a nice list of groups and the datasets found within them using the variable
        ``self.paths``.

        Example:
            Print any groups found within the HDF5 file and the datasets within:

            >>> print(H5Store)
        """
        printout = (' ' * 21) + 'Groups and datasets found in HDF5 file\n' + ('-' * 80) + '\n'
        for group in sorted(self.paths):
            printout = printout + ('Group Path: \'%s\'' % group) + '\n'
            printout = printout + textwrap.fill('Datasets:   %s' % self.paths[group], width=80,
                                                subsequent_indent='             ') + '\n'
            printout = printout + ('-' * 80) + '\n'

        return printout

    def __getitem__(self, key):
        """
        Retrieve an array using the self[key] notation. The passed key is the path used to access
        the array desired and included direction through groups if required and the dataset name.
        The array may be selectively queried allowing a specific value or range of values to be
        loaded into the systems memory and not the whole array.

        Example:

            Retrieve a single 1-D array:

            >>> array = H5Store['Group1/Group2/Var1']

            Retrieve a single 1-D array within the root directory of the HDF5 file:

            >>> array = H5Store['Var1']

            Retrieve the first value from the array:

            >>> value = H5Store['Var1', 0]

            Retrieve a slice of values from the array:

            >>> values = H5Store['Var1', 10:15]
        """

        # Manage the input
        rng = None
        if isinstance(key, tuple):
            key, rng = key
        if key in ['', ' ']:
            key = '/'
        if not key.endswith("/"):
            key += "/"

        # Sanity checks
        if self.h5data is None:
            raise FileNotFoundError("No HDF5 file exists yet")
        if rng is not None:
            if not isinstance(self.h5data[key], self.h5py.Dataset):
                raise KeyError("The passed key must lead to a H5 dataset if a slice is desired")

        # Load the full group or just a single array
        if isinstance(self.h5data[key], self.h5py.Group):
            datasets = [x for x in list(self.h5data[key].keys()) if
                        isinstance(self.h5data['%s%s' % (key, x)], self.h5py.Dataset)]
            data = []
            for dataset in datasets:
                data.append(self.h5data['%s/%s' % (key, dataset)][:].flatten())
            return pd.DataFrame(np.column_stack(data), columns=datasets)
        elif rng is not None:
            return self.h5data[key][rng]
        else:
            return self.h5data[key][:]

    def __setitem__(self, key, value):
        """
        Write the the HDF5 file using the self[key] notation.

        If a pd.Series or np.array is passed, the last entry in the path is used as the dataset
        name. If a ``pd.DataFrame`` is passed, all columns are written to the path specified to
        datasets with their names retrieved from the ``pd.DataFrame``'s columns. If more flexible
        usage is required, please use :func:`gs.write_h5()<pygeostat.data.h5_io.write_h5>`.

        Example:
            Write a np.array or pd.Series to the HDF5 file:

            >>> H5Store['Group1/Group2/Var1'] = np.array()

            Write all the columns in a ``pd.DataFrame`` to the HDF5 file:

            >>> H5Store['Group1/Group2'] = pd.DataFrame()
        """

        # Handle different types of input objects
        if isinstance(value, pd.DataFrame):
            h5path = key
            variables = None
        elif isinstance(value, pd.Series):
            h5path = key
            variables = [value.name]
        else:
            key = key.rsplit('/', 1)
            if len(key) > 1:
                h5path, variables = key
            else:
                h5path = None
                variables = key[0]

        # Write the data to the HDF5 file
        write_h5(value, self.flname, h5path, variables)

        # Reload the file
        if isinstance(self.h5data, self.h5py.File):
            self.h5data.close()
        self._loadfile()

    def __enter__(self):
        " required for `with gs.H5Store(file) as h5store:` pattern "
        return self

    def __exit__(self, type, value, traceback):
        " required for `with gs.H5Store(file) as h5store:` pattern "
        self.close()

    def _loadfile(self):
        """
        Load the HDF5 file as h5py object and get the dataset paths
        """
        self.h5data = self.h5py.File(self.flname, 'r+')
        self.paths = self._get_paths()

    def _get_paths(self):
        """
        Build lookup dictionary of all the paths to the datasets found in a HDF5 file.
        """
        done = False
        paths = {}
        groups = ['/']
        while not done:
            for group in groups:
                children = ['%s%s/' % (group, x)for x in list(self.h5data[group].keys()) if
                            isinstance(self.h5data['%s%s' % (group, x)], self.h5py.Group)]
                groups.extend(children)
                datasets = [x for x in list(self.h5data[group].keys()) if
                            isinstance(self.h5data['%s%s' % (group, x)], self.h5py.Dataset)]
                if len(datasets) > 0:
                    paths[group] = datasets
                groups.remove(group)
            if len(groups) == 0:
                done = True

        return paths

    def close(self):
        """
        Release the open HDF5 file from python.
        """
        try:
            self.h5data.close()
        except:
            print("The H5 file wasn't open")

    def datasets(self, h5path=None):
        """
        Return the datasets found in the specified group.

        Keyword Arguments:
            h5path (str): Forward slash (/) delimited path through the group hierarchy you wish to
                retrieve the lists of datasets from. A dataset name cannot be passed using this
                argument, it is interpreted as a group name. A value of ``None`` places the dataset
                into the root directory of the HDF5 file.

        Returns:
            datasets (list): List of the datasets found within the specified `h5path`
        """
        h5path = _fixh5path(h5path)
        return self.paths[h5path]

    def iteritems(self, h5path=None, datasets=None, wildcard=None):
        """
        Produces an iterator that can be used to iterate over HDF5 datasets.

        Can use the parameter ``h5path`` to indicate which group to retrieve the datasets from.
        If a set of specific datasets are required, the parameter ``datasets`` will restrict the
        iterator to those. The parameter ``wildcard`` allows a string wild-card value to restrict
        which datasets are iterated over.

        Keyword Arguments:
            h5path (str): Forward slash (/) delimited path through the group hierarchy you wish to
                retrieve datasets from. A dataset name cannot be passed using this argument, it is
                interpreted as a group name. A value of ``None`` places the dataset into the root
                directory of the HDF5 file.
            datasets (list): List of specific dataset names found within the specified group to
                iterator over
            wildcard (str): String to search for within the names of the datasets found within the
                specified group to iterate over

        Examples:

            Load a HDF5 file to pygeostat:

            >>> data = gs.H5Store('data.h5')

            Iterate over all datasets within the root directory of a HDF5 file:

            >>> for dataset in data.iteritems():
            >>>     gs.histplt(dataset)

            Iterate over the datasets within a specific group that are realizations:

            >>> for dataset in data.iteritems(h5path='Simulation/NS_AU', wildcard='Realization'):
            >>>     gs.histplt(dataset)
        """
        # Handle some parameters
        h5path = _fixh5path(h5path)
        if isinstance(datasets, str):
            datasets = [datasets]
        # Sanity checks
        if datasets and wildcard:
            raise ValueError("Only `datasets` or `wildcard` can be passed")
        if datasets:
            error = set(datasets) - set(self.paths[h5path])
            if len(error) > 0:
                raise ValueError("The following datasets were specified but not found in the H5"
                                 " group: %s" % error)
        elif wildcard:
            datasets = []
            for dataset in self.paths[h5path]:
                if wildcard in dataset:
                    datasets.append(dataset)
            if len(datasets) == 0:
                raise ValueError("No datasets with the specified `wildcard` were found")
        else:
            datasets = self.paths[h5path]
            if len(datasets) == 0:
                raise ValueError("No datasets were found in the specified `h5path`")
        # Produce the generator
        for dataset in datasets:
            yield self.__getitem__(dataset)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Generate a correlation matrix plot"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------


from scipy import cluster
import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
from . set_style import set_plot_style
from .. pygeostat_parameters import Parameters


@set_plot_style
def correlation_matrix_plot(correlation_data, figsize=None, ax=None, cax=None, title=None, xticklabels=None,
                            ticklabels=None, yticklabels=None, rotateticks=None, cbar=None, annotation=None, lower_matrix=False,
                            lw=0.5, hierarchy=None, dendrogram=False, vlim=(-1, 1), cbar_label=None, cmap=None,
                            plot_style=None, custom_style=None, output_file=None, out_kws=None, sigfigs=3, **kwargs):
    """
    This function uses matplotlib to create a correlation matrix heatmap illustrating the
    correlation coefficient between each pair of variables.

    The only parameter needed is the correlation matrix. All of the other arguments are optional.
    Figure size will likely have to be manually adjusted. If the label parameters are left to their
    default value of ``None`` and the input matrix is contained in a pandas dataframe, the
    index/column information will be used to label the columns and rows. If a numpy array is
    passed, axis tick labels will need to be provided. Axis tick labels are automatically checked
    for overlap and if needed, are rotated. If rotation is necessary, consider condensing the
    variables names or plotting a larger figure as the result is odd. If ``cbar`` is left to its
    default value of ``None``, a colorbar will only be plotted if the ``lower_matrix`` is set to True. It
    can also be turned on or off manually. If ``annotation`` is left to its default value of ``None``,
    annotations will only be placed if a full matrix is being plotted. It can also be turned on or
    off manually.

    The parameter ``ticklabels`` is odd in that it can take a few forms, all of which are a tuple
    with the first value controlling the x-axis and second value controlling the y-axis (x, y). If
    left to its default of ``None``, another pygeostat function will check to see if the labels
    overlap, if so it will rotate the axis labels by a default angle of (45, -45) if required. If
    a value of ``True`` is pased for either axis, the respective default values previously stated
    is used. If either value is a float, that value is used to rotate the axis labels.

    The correlation matrix can be ordered based on hierarchical clustering. The following is a list
    of permissible arguments: ``'single', 'complete', 'average', 'weighted', 'centroid', 'median',
    'ward'``. The denrogram if plotted will have a height equal to 15% the height of the
    correlation matrix. This is currently hard coded.

    .. seealso::
        http://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html#scipy.cluster.hierarchy.linkage

    Please review the documentation of the :func:`gs.set_style()
    <pygeostat.plotting.set_style.set_style>` and :func:`gs.export_image()
    <pygeostat.plotting.export_image.export_image>` functions for details on their parameters so that
    their use in this function can be understood.

    Parameters:
        correlation_data: Pandas dataframe or numpy matrix containing the required loadings or correlation
            matrix
        figsize (tuple): Figure size (width, height)
        ax (mpl.axis): Matplotlib axis to plot the figure
        title (str): Title for the plot
        ticklabels (list): Tick labels for both axes
        xticklabels (list): Tick labels along the x-axis (overwritten if ticklabels is passed)
        yticklabels (list): Tick labels along the y-axis (overwritten if ticklabels is passed)
        rotateticks (bool or float tuple): Bool or float values to control axis label rotations.
            See above for more info.
        cbar (bool): Indicate if a colorbar should be plotted or not
        annotation (bool): Indicate if the cells should be annotationated or not
        lower_matrix (bool): Indicate if only the lower matrix should be plotted
        lw (float): Line width of lines in correlation matrix
        hierarchy (str): Indicate the type of hieriarial clustering to use to reorder the
            correlation matrix. Please see above for more details
        dendrogram (bool): Indicate if a dendrogram should be plotted. The argument ``hierarchy``
            must be set to ``true`` for this argument to have any effect
        vlim (tuple): vlim for the data on the correlation_matrix_plot, default = (-1, 1)
        cbar_label (str): string for the colorbar label
        cmap (str): valid Matplotlib colormap
        plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
            :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
            to turn it off
        custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
        output_file (str): Output figure file name and location
        out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
            :func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`
        sigfigs (int): significant digits for labeling of colorbar and cells
        **kwargs: Optional permissible keyword arguments to pass to matplotlib's pcolormesh
            function

    Returns:
        ax (ax): matplotlib Axes object with the correlation matrix plot

    **Examples:**
        Calculate the correlation matrix variables in a pandas dataframe

        .. plot::
        
            import pygeostat as gs
            data_file = gs.ExampleData("point3d_ind_mv")
            data = data_file[data_file.variables]
            data_cor = data.corr()
            gs.correlation_matrix_plot(data_cor, cmap = 'bwr')

        |

        Again for illustration, convert the correlation dataframe into a numpy matrix. By using a
        numpy matrix, the axis labels will need to me manually entered. Reduce the figure size as
        well:

        .. plot::
        
            import pygeostat as gs
            data_file = gs.ExampleData("point3d_ind_mv")
            data = data_file[data_file.variables]
            data_cor = data.corr()
            gs.correlation_matrix_plot(data_cor.values, cmap = 'bwr')

        |

        Plotting a lower correlation matrix while having
        annotations:

        .. plot::

            import pygeostat as gs

            data_file = gs.ExampleData("point3d_ind_mv")
            data = data_file[data_file.variables]
            data_cor = data.corr()
            gs.correlation_matrix_plot(data_cor, lower_matrix=True, annotation=True)
                
    """
  
    from . export_image import export_image
    from . utils import titleoverlap, _tickoverlap, get_contcbarargs
    from mpl_toolkits.axes_grid1 import ImageGrid
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    from mpl_toolkits.axes_grid1 import axes_divider

    # Sanity checks
    if lower_matrix and dendrogram:
        raise NotImplementedError("Dendrogram plotting while using the ``lower_matrix`` functionality is"
                                  " not currently supported")
    # Convert the numpy array to a pd.DataFrame (needed for hierarchy)
    if isinstance(correlation_data, np.ndarray):
        correlation_data = pd.DataFrame(data=correlation_data)

    if not isinstance (correlation_data, pd.DataFrame):
        raise ValueError('correlation_data must be convertable to pandas dataframe')

    # Set-up some parameters
    nx = correlation_data.shape[1]
    ny = correlation_data.shape[0]
    #  Handle dictionary defaults
    if out_kws is None:
        out_kws = dict()
    # Plot Colourbar if required
    if cbar is None:
        if lower_matrix or annotation is False:
            cbar = True
        else:
            cbar = False
    #  Determine hierarchy if needed
    if hierarchy in ['single', 'complete', 'average', 'weighted', 'centroid', 'median', 'ward']:
        # Determine the clustering
        linkage = cluster.hierarchy.linkage(correlation_data, method=hierarchy)
        dendo = cluster.hierarchy.dendrogram(linkage, no_plot=True)
        hierarchy = True
        # Reorder the correlation matrix and the tick labels
        ind = dendo['leaves']
        xlab = list(correlation_data.columns)
        xorder = [xlab[i] for i in ind]
        correlation_data = correlation_data[xorder]
        ylab = list(correlation_data.index)
        yorder = [ylab[i] for i in ind]
        correlation_data = correlation_data.loc[yorder]
        if ticklabels:
            ticklabels = [ticklabels[i] for i in ind]
        if xticklabels:
            xticklabels = [xticklabels[i] for i in ind]
        if yticklabels:
            yticklabels = [yticklabels[i] for i in ind]
    # Copy the data and convert it to a numpy matrix if a pandas dataframe was passed
    if isinstance(correlation_data, pd.DataFrame):
        plot_data = correlation_data.values
    else:
        plot_data = np.asarray(correlation_data)
    # Set-up plot else coerce the passed axes as required
    if ax is None:
        if cbar:
            if hierarchy:
                cbar_mode = 'each'
            else:
                cbar_mode = 'single'
        else:
            cbar_mode = None
        if hierarchy and dendrogram:
            nrows_ncols = (2, 1)
        else:
            nrows_ncols = (1, 1)
        # Setup up a new plot
        fig = plt.figure(figsize=figsize)
        imggrid = ImageGrid(fig, 111, nrows_ncols, axes_pad=(0.07, 0), cbar_mode=cbar_mode, cbar_size=0.075)
        ax = imggrid[0]
        if hierarchy and dendrogram:
            ax_dendo = imggrid[1]
            fig.delaxes(imggrid.cbar_axes[1])
        if cbar:
            cax = imggrid.cbar_axes[0]
    elif (cbar or hierarchy) and not isinstance(ax, axes_divider.LocatableAxes):
        fig = plt.gcf()
        divider = make_axes_locatable(ax)
        if cbar:
            cax = divider.append_axes("right", size=0.075, pad=0.07, aspect='auto')
        if hierarchy and dendrogram:
            ax_dendo = divider.append_axes("bottom", size=0.4, pad=0.0, aspect='auto')
    elif hierarchy:
        raise ValueError("`ax` cannotation be divided meaning the dendrogram cannotation be plotted")
    elif cbar and cax is None:
        raise ValueError("A colorbar axes `cax` must be passed as the passed `ax` cannotation be"
                         " divided.")
    ax.set_aspect('equal')
    # Set the axis ticklabels if possible
    if ticklabels:
        xlabels = ticklabels
        ylabels = ticklabels
    else:
        # Handle xticklabels
        if xticklabels is None and hasattr(correlation_data, 'columns'):
            xticklabels = list(correlation_data.columns)
        elif isinstance(xticklabels, bool) and xticklabels:
            xticklabels = list(correlation_data.columns)
        if xticklabels is not None:
            xlabels = xticklabels
        else:
            xlabels = None
        # Handle yticklabels
        if yticklabels is None and hasattr(correlation_data, 'index'):
            yticklabels = list(correlation_data.index)
        elif isinstance(yticklabels, bool) and yticklabels:
            yticklabels = list(correlation_data.index)
        if yticklabels is not None:
            ylabels = yticklabels
        else:
            ylabels = None
    # Set-up figure estetics
    if lower_matrix:
        gridclr = 'white'
        ax.set(xlim=(0, nx - 1), ylim=(0, ny - 1))
        xticklocs = np.arange(nx - 1)
        yticklocs = np.arange(ny - 1)
        # Trim the labels
        if isinstance(ylabels, list):
            ylabels = ylabels[1:]
    else:
        gridclr = 'black'
        ax.xaxis.tick_top()
        ax.set(xlim=(0, nx), ylim=(0, ny))
        xticklocs = np.arange(nx)
        yticklocs = np.arange(ny)
    # Set-up x-axis labels and grid locations
    ax.set_xticks(xticklocs + 0.5)
    ax.set_xticks(xticklocs, minor=True)
    if xlabels is not None:
        if lower_matrix:
            va = 'top'
        else:
            va = 'bottom'
        ax.set_xticklabels(xlabels, va=va, ha='center', rotation='horizontal')
        ax.tick_params(axis='x', pad=2)
    else:
        ax.get_xaxis().set_ticks([])
    # Set-up y-axis labels and grid locations
    ax.invert_yaxis()
    ax.set_yticks(yticklocs + 0.5)
    ax.set_yticks(yticklocs, minor=True)
    if ylabels is not None:
        ax.set_yticklabels(ylabels, va='center', ha='right', rotation='vertical')
        ax.tick_params(axis='y', pad=1)
    else:
        ax.get_yaxis().set_ticks([])
    # Set-up the figure spines
    for spine in ax.spines:
        if lower_matrix:
            ax.spines[spine].set_color('white')
    # Mask the data if lower_matrix is being used
    if lower_matrix:
        mask = np.zeros_like(correlation_data, dtype=np.bool)
        mask[np.triu_indices_from(mask)] = True
        plot_data = np.ma.masked_where(mask, plot_data)
        plot_data = plot_data[1:, :(ny - 1)]
    # Plot the grid
    ax.grid(True, which='minor', color=gridclr, zorder=3, lw=lw)
    # Tick rotation
    plt.draw()
    # Check if the axis tick labels overlap, if so rotate them
    if rotateticks is None:
        rotateticks = Parameters['plotting.rotateticks']
    if rotateticks is None:
        # The plots tick labels will not be properly accessible until the figure is "drawn", once
        # the command below is run, ax.get_ticklabel() will actually work properly.
        rotateticks = _tickoverlap(ax)
    # Rotate if required
    if rotateticks[0] is not False or rotateticks[0] is not None:
        if rotateticks[0] is True:
            rotateticks[0] = 45
        xlabels = ax.get_xticklabels()
        for xlabel in xlabels:
            if lower_matrix:
                xlabel.set_ha('center')
                xlabel.set_va('top')
            else:
                xlabel.set_ha('center')
                xlabel.set_va('bottom')
            xlabel.set_rotation(rotateticks[0])
    if rotateticks[1] is not False or rotateticks[1] is not None:
        if rotateticks[1] is True:
            rotateticks[1] = -45
        ylabels = ax.get_yticklabels()
        for ylabel in ylabels:
            ylabel.set_ha('right')
            ylabel.set_va('center')
            ylabel.set_rotation(rotateticks[1])
        ax.tick_params(axis='y', pad=2)
    # Plot the title if required
    if title:
        titletxt = ax.set_title(title)
        # Due to placing the xticklaebls on the top, if a title is plotted it is going to overlap.
        # The following code checks for overlap and bumps the title up step by step until there
        # isn't any
        if not lower_matrix:
            # Check to see if there is overlap with the title and the xticklabels The plots tick
            # labels will not be properly accessible until the figure is "drawn", once the command
            # below is run, ax.titleoverlap() will actually work properly.
            plt.draw()
            _titleoverlap = titleoverlap(ax, titletxt)
            # If there is, clear the title and start moving a suptitle up until there isn't overlap
            if _titleoverlap:
                titletxt.set_text('')
                shifttitle = True
                y = 1.01
            else:
                shifttitle = False
            while _titleoverlap:
                titletxt = ax.set_title(title, y=y)
                plt.draw()
                _titleoverlap = titleoverlap(ax, titletxt)
                y = y + 0.01
            # Now that a spot without overlap has been found, add the pad of 0.015 (0.01 alread
            # added to y) so that it is slightly farther away from the axis labels
            if shifttitle:
                titletxt = ax.set_title(title, y=(y + 0.005))
    # Plot the figure
    if cmap is None:
        cmap = Parameters['plotting.cmap']
    plot = ax.pcolormesh(plot_data, cmap=cmap, norm=plt.Normalize(vmin=vlim[0], vmax=vlim[1]),
                         zorder=0)
    # annotationate if required
    if not isinstance(annotation, bool):
        if lower_matrix:
            annotation = False
        else:
            annotation = True
    if annotation:
        # Set-up a colormap to use to color the annotationations. This was manually tuned so if you can
        # come up with something better do it
        clrvals = [0.85, 0.85, 0.85, 0.85, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        if cmap is None:
            cmap = []
            for clr in clrvals:
                cmap.append(str(clr))
        if isinstance(cmap, list):
            cmap = mpl.colors.ListedColormap(cmap)
            clrnorm = mpl.colors.Normalize(vmin=vlim[0], vmax=vlim[1])
            clrmap = mpl.cm.ScalarMappable(norm=clrnorm, cmap=cmap)
        for y in range(plot_data.shape[0]):
            for x in range(plot_data.shape[1]):
                if isinstance(plot_data[y, x], np.float):
                    # try:
                    #     color = clrmap.to_rgba(plot_data[y, x])
                    # except (KeyError, ValueError):
                    color = 'black'
                    ax.text(x + 0.5, y + 0.5, ('{:.%ig}' % sigfigs).format(plot_data[y, x]),
                            ha='center', va='center', color=color)
    if cbar:
        # handle parms for colorbars and colormaps
        vlim, ticklocs, ticklabels = get_contcbarargs(np.linspace(vlim[0], vlim[1], 5),
                                                         sigfigs, vlim)
        # Plot the colorbar
        cbar = plt.colorbar(plot, cax=cax, ticks=ticklocs)
        # Configure the color bar
        cbar.ax.set_yticklabels(ticklabels, ha='left')
        cbar.ax.tick_params(axis='y', pad=2)
        if cbar_label is not None:
            cbar.set_label(cbar_label, ha='center', va='top', labelpad=2)
    # Plot the dendrogram if required
    if hierarchy and dendrogram:
        # Get the line coordinates and scale
        ylines = np.array(dendo['dcoord'])
        ydendo = ny * 0.15
        ylines = ylines * (ydendo / ylines.max())
        xlines = np.array(dendo['icoord']) / 10
        lines = []
        for (xline, yline) in zip(xlines, ylines):
            lines.append(list(zip(xline, yline)))
        # Plot the dendrogram
        coll = mpl.collections.LineCollection(lines, color='k', lw=lw)
        ax_dendo.add_collection(coll)
        # Fix the subplots aesthetics
        ax_dendo.set_ylim(ydendo, 0)
        for spine in ax_dendo.spines:
            ax_dendo.spines[spine].set_visible(False)
        ax_dendo.yaxis.set_visible(False)
        ax_dendo.xaxis.set_visible(False)
        ax_dendo.patch.set_visible(False)
    # Export figure
    if output_file or ('pdfpages' in out_kws):
        export_image(output_file, **out_kws)

    return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""image_grid.py: Provides a means to easily generate grids of plots"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
from . set_style import set_plot_style


@set_plot_style
def image_grid(ncol, nrow, symmetric=False, nvar=None, gridfunc=None, upperfunc=None,
			  lowerfunc=None, diagfunc=None, tight=False, axes_pad=0.2, cbar=False, vlim=None,
			  cbar_label=None, figsize=None, aspect=None, xlim=None, ylim=None, axislabels=None,
			  xlabel=None, ylabel=None, suptitle=None, ntickbins=2, rotateticks=False,
			  plot_style=None, custom_style=None, output_file=None, labelmode='all',
			  unequal_aspects=False, direction='row', **kwargs):
	"""
	Create either a symmetric or non-symmetric plot matrix. This function interprets a symmetric
	plot matrix as one that plots multivariate data in different types of bivariate plots in the
	lower and/or upper triangles and univariate plots along the diagonal. Conversely, a non-
	symmetric plot matrix assumes a single plotting function is used to populate all of the
	subplots.

	To provide a very flexible plotting skeleton, this function does not actually instruct any
	plotting functions. Instead, the user is required to define python generators that contain a
	loop that will produce the desired subplots. Use of generators allows the user to
	customize the subplots as desired, without having to use this function as a middle man.

	If plotting a symmetric plot matrix, the keyword argument ``nvar`` is required and one of the
	following is required: ``upperfunc``, ``owerfunc``, or ``diagfunc``. If plotting a
	non-symetric plot matrix, the following keyword arguments are required: ``nvar``, ``ncol``,and
	``gridfunc``.

	.. note::

		The subplots regardless of their location, plot left to right, top to bottom.

		The plots aspect will need to be considered. If your plots appear flat, your aspect is
		wrong. As an example, variograms are classically plotted at a 4:3 ratio, meaning if
		the y-axis limits are left to their default of 0 to 1.2 and say your x-axis is being
		plotted to 1000, you would require a aspect of :math:`{1000 / 1.2 / (4/3)}` or 62.5. If
		this function detects a y-axis limit of 1.2, it will calculate an aspect automatically
		unless ``aspect`` is set manually.

	Colorbars can be plotted for each individual subplot by setting keyword argument ``cbar`` to
	``'each'``; however, with this setting, this function does not handle any colorbar plotting.
	Therefore, the color bar axes are passed to the iterator and the plotting function within the
	iterator deals with the colorbar. If a single colorbar is desired, ``cbar`` is set to
	``'single'``. When using this setting, the keyword argument ``vlim`` must be passed. It is also
	important that all of the subplots have their colormaps limited to this same range. If no color
	bar(s) are desired, the keyword argument is set to ``'none'``.

	Please review the documentation of the
	:func:`gs.set_style() <pygeostat.plotting.set_style>` and
	:func:`gs.export_image() <pygeostat.plotting.export_image>` functions for details on their
	parameters so that their use in this function can be understood.

	Keyword Arguments:
		symmetric (bool):
		nvar (int): Only used for symmetric grids. Number of to variables to plot
		ncol (int): Only used for non-symmetric grids. Number of to columns to plot
		nrow (int): Only used for non-symmetric grids. Number of to rows to plot
		gridfunc (generator): Python generator that contains a loop that can be used to plot the
			desire subplots for the whole grid
		upperfunc (generator): Python generator that contains a loop that can be used to plot the
			desire subplots in the upper triangle of the grid
		lowerfunc (generator): Python generator that contains a loop that can be used to plot the
			desire subplots in the lower triangle of the grid
		diagfunc (generator): Python generator that contains a loop that can be used to plot the
			desire subplots along the diagonal of the grid
		tight (bool): Indicate if the whitespace between the subplots should be removed.
			Reminiscent of R's scatter
		axes_pad (float or tuple): Padding in iches to place between the plots. Can pass a tuple to
			indicate different padding in the horizontal and veritcal directions (width_pad,
			height_pad)
		cbar (str): Indicate what colorbar mode to use. The available options are ``['none',
			'single', 'each']``. See above for instructions
		vlim (tuple): If ``cbar`` is set to ``'single'``, this value is required and instructs the
			limits of the colorbar.
		cbar_label (str): Colorbar title
		figsize (tuple): Figure size (width, height)
		aspect (str): Set a permissible aspect ratio of the image to pass to matplotlib. The
			function will try and detect what aspect is best as described above.
		xlim (float tuple): X-axis limits applied to all axes in the grid
		ylim (float tuple): Y-axis limits applied to all axes in the grid
		axislabels (list): Only used for symmetric grids. Labels for each row and column
		xlabel (str): Super x-axis label
		yalabl (str): Super y-axis label
		suptitle (str): Super title
		ntickbins (int or tuple): int: applied to both x and y, or tuple, applied to x and y
			respectively
		rotateticks (bool or float tuple): Bool or float values to control axis label rotations.
			See above for more info.
		plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
			:class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
			to turn it off
		custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
		output_file (str): Output figure file name and location
		labelmode (str): Labeling input to the image_grid function. Default is ``'all'``. ``'L'``
			labels the left column and bottom row only. There may be other valid parameters. The
			last plot in each column will get x-axis labels in ``'L'`` mode as unused subplots in
			the last row are removed. If ``tight`` is set to ``True``, ``'L'`` is always used.
		unequal_aspects (bool): Indicate `True` if the limits of the plots will differe between
			subplots. This will then use plt.subplots() rather than image_grid() for plotting
			functions
		**kwargs: Optional permissible keyword arguments to pass to
			:func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`

	Examples:

		Below are a few images illustrating the results of this function.

		Accuracy plot matrix using the results from multivariate simulation:

		.. image:: ./figures/advancedplotting_imggrid_accplt.png

	|
	"""
	import types
	import numpy as np
	from .export_image import export_image
	from .utils import get_contcbarargs, get_supaxislocs
	import matplotlib as mpl
	import matplotlib.pyplot as plt
	from mpl_toolkits.axes_grid1 import ImageGrid, LocatableAxes, SubplotDivider

	# Sanity checks
	if symmetric:
		if nvar is None:
			raise ValueError("When plotting a symmetric grid, the keyword argument `nvar` must be"
							 " passed")
		if all([upperfunc is None, lowerfunc is None, diagfunc is None]):
			raise ValueError("When plotting a symmetric grid, one of the keyword arguments"
							 "`upperfunc`, `lowerfunc`, or `diagfunc` must be passed")
	else:
		if any([ncol is None, nrow is None]):
			raise ValueError("When plotting a non-symmetric grid, the keyword arguments `ncol` and"
							 " `nrow` must be pased")
		if gridfunc is None:
			raise ValueError("When plotting a non-symmetric grid, the keyword argument `gridfunc`"
							 " must be passed")
	# ------------------------------------------------
	# Setup Styles, Parameters, and Image Grid
	# ------------------------------------------------
	# Sort out some figure parameters
	if symmetric:
		if ncol is None:
			ncol = nvar
		if nrow is None:
			nrow = nvar
	if tight:
		labelmode = 'L'
		axes_pad = 0
		fontsize = mpl.rcParams['font.size']
		if fontsize > 8:
			scalar = (0.42 * (fontsize / 16)) + 1
		elif fontsize < 8:
			scalar = (0.83 * (fontsize / 5))
		else:
			scalar = 1
		cbar_pad = 0.2 + 0.18 * scalar
		cbar_location = 'bottom'
	else:
		cbar_pad = None
		cbar_location = 'right'
	if cbar in ['single', 'each']:
		if cbar == 'each':
			cbar_pad = 0.05
			cbar_size = 0.075
		else:
			cbar_size = 0.075
	else:
		cbar_size = None
	# ------------------------------------------------
	# Setup the axes either using plt.subplots() or image_grid based on `unequal_aspects`
	# ------------------------------------------------
	if unequal_aspects:
		# checks for unhandled combinations
		if tight:
			raise ValueError('`tight=True` and `unequal_aspects=True` cannot be used together '
							 'with `gs.image_grid()`')
		if cbar:
			raise ValueError('Cannot specify `cbar` with `unequal_aspects=True`, use the `cbar` '
							 'plotting options in the respective iterator functions instead')
		# setup this function to use plt.subplots
		fig, grid = plt.subplots(nrow, ncol, figsize=figsize, facecolor='none')
		ax_all = grid.flatten().tolist()
		if isinstance(axes_pad, tuple):
			plt.subplots_adjust(wspace=axes_pad[0], hspace=axes_pad[1])
		elif isinstance(axes_pad, (int, float)):
			plt.subplots_adjust(wspace=axes_pad, hspace=axes_pad)
		else:
			plt.tight_layout(pad=0.4, w_pad=0.5, h_pad=1.0)
	else:
		# Set-up Figure object and image grid
		fig = plt.figure(figsize=figsize)
		grid = ImageGrid(fig, 111, nrows_ncols=(nrow, ncol), axes_pad=axes_pad, cbar_mode=cbar,
						 cbar_size=cbar_size, cbar_pad=cbar_pad, cbar_location=cbar_location,
						 share_all=False, label_mode=labelmode, direction=direction)
		ax_all = grid.axes_all
	# Fix some figure aesthetics prior to the plotting function so the user is able to move away
	# from these defaults if desired
	for ax in ax_all:
		if isinstance(ntickbins, tuple):
			ax.locator_params(axis='x', nbins=ntickbins[0])
			ax.locator_params(axis='y', nbins=ntickbins[1])
		else:
			ax.locator_params(nbins=ntickbins)
	# Get the axes for the lower, upper, and diagonal
	if symmetric:
		indexes = np.arange(nvar * nvar).reshape(nvar, nvar)
		lower = np.tril_indices(nvar, k=-1)
		upper = np.triu_indices(nvar, k=1)
		diag = np.diag_indices(nvar)
		ax_lower = [ax_all[iax] for iax in indexes[lower].tolist()]
		ax_upper = [ax_all[iax] for iax in indexes[upper].tolist()]
		ax_diag = [ax_all[iax] for iax in indexes[diag].tolist()]
	cax_lower = cax_upper = cax_diag = []
	if cbar == 'each':
		if symmetric:
			args = tuple([grid.cbar_axes])
			cax_lower = [grid.cbar_axes[iax] for iax in indexes[lower].tolist()]
			cax_upper = [grid.cbar_axes[iax] for iax in indexes[upper].tolist()]
			cax_diag = [grid.cbar_axes[iax] for iax in indexes[diag].tolist()]
		else:
			args = tuple([grid.cbar_axes])
	else:
		args = tuple()
	# ------------------------------------------------
	# Plot data using operator functions
	# ------------------------------------------------
	if symmetric:
		if isinstance(lowerfunc, types.FunctionType):
			for ax in lowerfunc(ax_lower, *args):
				continue
		else:
			for ax in ax_lower:
				fig.delaxes(ax)
			for cax in cax_lower:
				fig.delaxes(cax)
		if isinstance(diagfunc, types.FunctionType):
			for ax in diagfunc(ax_diag, *args):
				continue
		else:
			for ax in ax_diag:
				fig.delaxes(ax)
			for cax in cax_diag:
				fig.delaxes(cax)
		if isinstance(upperfunc, types.FunctionType):
			for ax in upperfunc(ax_upper, *args):
				continue
		else:
			for ax in ax_upper:
				fig.delaxes(ax)
			for cax in cax_upper:
				fig.delaxes(cax)
	else:
		nplot = 0
		for ax in gridfunc(ax_all, *args):
			nplot += 1
	# ------------------------------------------------
	# Fix some figure aesthetics after plotting
	# ------------------------------------------------

	if aspect is None:
		ax = ax_all[0]
		pltxlim = ax.get_xlim()[1]
		pltylim = ax.get_ylim()[1]
		# Sort out a 4:3 ratio for variograms
		if pltylim == 1.2:
			aspect = pltxlim / 1.2 / (4 / 3)
		# or the 4:3 ratio for CDF's
		elif pltylim == 1.0:
			aspect = pltxlim / 1.0 / (4 / 3)
	for ax in ax_all:
		ax.tick_params(axis='both', pad=2)
		if xlim:
			ax.set_xlim(xlim)
		if ylim:
			ax.set_ylim(ylim)
		if aspect:
			ax.set_aspect(aspect)
		if tight:
			for spine in ax.spines:
				ax.spines[spine].set_visible(True)
		# Set-up the tick labels
		if symmetric and (labelmode.lower() == 'l' or tight):
			ax.set_xlabel('')
			ax.set_ylabel('')
			# y-axis labels
			if isinstance(lowerfunc, types.FunctionType):
				if tight:
					# Handle left labels
					if isinstance(diagfunc, types.FunctionType):
						for iax in indexes[:, 0][::2]:
							ax = ax_all[iax]
							ax.axis['left'].toggle(ticklabels=True, label=True)
						for iax in indexes[:, 0][1::2]:
							ax = ax_all[iax]
							ax.axis['left'].toggle(ticklabels=False, label=False)
					else:
						for iax in indexes[1:, 0][::2]:
							ax = ax_all[iax]
							ax.axis['left'].toggle(ticklabels=True, label=True)
						for iax in indexes[1:, 0][1::2]:
							ax = ax_all[iax]
							ax.axis['left'].toggle(ticklabels=False, label=False)
					# Handle right labels
					if isinstance(diagfunc, types.FunctionType):
						for ax in ax_diag[1::2]:
							ax.axis['right'].toggle(ticklabels=True, label=True)
					elif isinstance(upperfunc, types.FunctionType):
						for iax in indexes[1:, nvar - 1][1::2]:
							ax = ax_all[iax]
							ax.axis['right'].toggle(ticklabels=True, label=True)
					else:
						for iax in indexes[diag][:-1][::2] - 1:
							ax = ax_all[iax]
							ax.axis['right'].toggle(ticklabels=True, label=True)
				else:
					for iax in indexes[:, 0]:
						ax = ax_all[iax]
						ax.axis['left'].toggle(ticklabels=True, label=True)
			elif isinstance(upperfunc, types.FunctionType):
				if tight:
					# Handle right labels
					for iax in indexes[:, -1][::2]:
						ax = ax_all[iax]
						ax.axis['right'].toggle(ticklabels=True, label=True)
					# Handle left labels
					if isinstance(diagfunc, types.FunctionType):
						for ax in ax_diag[1::2]:
							ax.axis['left'].toggle(ticklabels=True, label=True)
					else:
						for iax in indexes[diag][:-1][1::2] + 1:
							ax = ax_all[iax]
							ax.axis['left'].toggle(ticklabels=True, label=True)
				else:
					if isinstance(diagfunc, types.FunctionType):
						for ax in ax_diag:
							ax.axis['left'].toggle(ticklabels=True, label=True)
					else:
						for iax in indexes[diag][:-1] + 1:
							ax = ax_all[iax]
							ax.axis['left'].toggle(ticklabels=True, label=True)
			else:
				for ax in ax_diag:
					ax.axis['left'].toggle(ticklabels=True, label=True)
			# x-axis labels
			if isinstance(lowerfunc, types.FunctionType):
				if tight:
					# Handle bottom labels
					for iax in indexes[-1, :][::2]:
						ax = ax_all[iax]
						ax.axis['bottom'].toggle(ticklabels=True, label=True)
					for iax in indexes[-1, :][1::2]:
						ax = ax_all[iax]
						ax.axis['bottom'].toggle(ticklabels=False, label=False)
					# Handle top labels
					if isinstance(diagfunc, types.FunctionType):
						for ax in ax_diag[1::2]:
							ax.axis['top'].toggle(ticklabels=True, label=True)
					elif isinstance(upperfunc, types.FunctionType):
						for iax in indexes[1:, nvar - 1][1::2]:
							ax = ax_all[iax]
							ax.axis['top'].toggle(ticklabels=True, label=True)
					else:
						for iax in indexes[diag][:-1][::2] - 1:
							ax = ax_all[iax]
							ax.axis['top'].toggle(ticklabels=True, label=True)
				else:
					for iax in indexes[-1, :]:
						ax = ax_all[iax]
						ax.axis['bottom'].toggle(ticklabels=True, label=True)
			elif isinstance(upperfunc, types.FunctionType):
				if tight:
					# Handle top labels
					if isinstance(diagfunc, types.FunctionType):
						for iax in indexes[0, :][::2]:
							ax = ax_all[iax]
							ax.axis['top'].toggle(ticklabels=True, label=True)
					else:
						for iax in indexes[0, 1:][::2]:
							ax = ax_all[iax]
							ax.axis['top'].toggle(ticklabels=True, label=True)
					# Handle bottom labels
					if isinstance(diagfunc, types.FunctionType):
						for iax in indexes[diag][:-1][::2] + 1:
							ax = ax_all[iax]
							ax.axis['bottom'].toggle(ticklabels=True, label=True)
					else:
						for iax in indexes[diag][:-1][1::2] + 1:
							ax = ax_all[iax]
							ax.axis['bottom'].toggle(ticklabels=True, label=True)
				else:
					if isinstance(diagfunc, types.FunctionType):
						for ax in ax_diag:
							ax.axis['bottom'].toggle(ticklabels=True, label=True)
					else:
						for iax in indexes[diag][:-1] + 1:
							ax = ax_all[iax]
							ax.axis['bottom'].toggle(ticklabels=True, label=True)
			else:
				for ax in ax_diag:
					ax.axis['bottom'].toggle(ticklabels=True, label=True)
		elif not symmetric and tight:
			if ((ncol * nrow) - nplot) % 2 == 0:
				icol = 0
			else:
				icol = 1
			for ax in grid.axes_column[ncol - 1][1::2]:
				ax.yaxis.set_visible(True)
				ax.yaxis.set_ticks_position('right')
			for ax in grid.axes_column[0][1::2]:
				ax.axis['left'].toggle(ticklabels=False, label=False)
			for ax in grid.axes_row[0][icol::2]:
				ax.xaxis.set_ticks_position('top')
				ax.xaxis.set_visible(True)
			for ax in grid.axes_row[nrow - 1][icol::2]:
				ax.axis['bottom'].toggle(ticklabels=False, label=False)
	plt.draw()
	# ------------------------------------------------
	# Remove unused plots and move the x-axis tick labels are required for non-symmetric plots
	# ------------------------------------------------
	if not symmetric:
		if nplot < (nrow * ncol):
			for i in range(nplot, (nrow * ncol)):
				# Delete the unused plot
				row, col = divmod(i, ncol)
				delaxes = False   # True == default..but perhaps they shouldnt be deleted?
				if delaxes:
					if hasattr(ax_all[row * ncol + col], 'cax'):
						fig.delaxes(ax_all[row * ncol + col].cax)
					fig.delaxes(ax_all[row * ncol + col])
				else:
					if hasattr(ax_all[row * ncol + col], 'cax'):
						fig.axes[row * ncol + col].cax.set_axis_off()
					fig.axes[row * ncol + col].set_axis_off()
				# Plot the x-axis on the plots above the one being deleted
				row = row - 1
				ax = ax_all[row * ncol + col]
				if not unequal_aspects:
					ax.axis["bottom"].toggle(ticklabels=True, label=True)
				if tight:
					for ax in grid.axes_row[nrow - 2][icol::2]:
						ax.axis['bottom'].toggle(ticklabels=False)
	# remove the labels if the unequal_aspects keyword and labelmode='l'
	if labelmode.lower() == 'l' and unequal_aspects:
		for iplot, ax in enumerate(ax_all):
			irow = np.floor(iplot / ncol)
			icol = iplot - irow * ncol
			if icol == 0 and irow < (nrow - 1):
				ax.set_xticklabels([])
				ax.set_xlabel('')
			elif irow == (nrow - 1) and icol > 0:
				ax.set_yticklabels([])
				ax.set_ylabel('')
			elif icol > 0 and irow < (nrow - 1):
				ax.set_xticklabels([])
				ax.set_yticklabels([])
				ax.set_xlabel('')
				ax.set_ylabel('')
	# ------------------------------------------------
	# Add super axis labels if not using the symmetric variable matrix or variable labels if so
	# ------------------------------------------------
	if symmetric and axislabels:
		# Add the axis labels so they are inline...for some reason the bbox of the ylabels are not
		#  providing the right coordinates so a different method is needed. This will not work if
		#  the bar is put on the left side.
		if tight:
			# Add xaxis labels
			ax = grid.axes_llc
			ax.set_xlabel(axislabels[0])
			ax.set_ylabel(axislabels[nvar - 1])
			renderer = fig.canvas.get_renderer()
			plt.draw()
			label = ax.xaxis.get_label()
			bbox = label.get_window_extent(renderer)
			ymax = bbox.ymax
			ymax = ax.transAxes.inverted().transform((0, ymax))[1]
			for i, ax in enumerate(grid.axes_row[nvar - 1]):
				ax.text(0.5, ymax, axislabels[i], ha='center', va='top', transform=ax.transAxes)
			# Add yaxis labels
			xmin, xmid, ymin, ymid, ymax = get_supaxislocs(fig, nvar, nvar, figsize, 0.01)
			for i, ax in enumerate(grid.axes_column[0]):
				bbox = ax.get_position()
				ax_ymax = bbox.ymax
				ax_ymin = bbox.ymin
				ax_ymid = ((ax_ymax - ax_ymin) / 2) + ax_ymin
				fig.text(xmin, ax_ymid, axislabels[i], rotation=90, ha='right', va='center')
		else:
			for i, ax in enumerate(grid.axes_column[0]):
				ax.set_ylabel(axislabels[i])
			for i, ax in enumerate(grid.axes_row[nvar - 1]):
				ax.set_xlabel(axislabels[i])
	elif (ylabel is not None) or (xlabel is not None) or (suptitle is not None):
		# Add superaxis titles
		for ax in ax_all:
			if xlabel is not None:
				ax.set_xlabel('')
			if ylabel is not None:
				ax.set_ylabel('')
		xmin, xmid, ymin, ymid, ymax = get_supaxislocs(fig, nrow, ncol, figsize, 0.01)
		if ylabel is not None:
			fig.text(xmin, ymid, ylabel, ha='right', va='center', rotation=90)
		if xlabel is not None:
			fig.text(xmid, ymin, xlabel, va='top', ha='center')
		if suptitle is not None:
			fig.text(xmid, ymax, suptitle, va='bottom', ha='center')

	# ------------------------------------------------
	# Mess with the colorbar
	# ------------------------------------------------

	if cbar == 'single':
		# Check to see if the limits of each plot are the same. May only work for pixelplt
		try:
			vlimcheck = False
			for ax in ax_all:
				if not vlimcheck:
					vmin = ax.get_images()[0].norm.vmin
					vmax = ax.get_images()[0].norm.vmax
					vlimcheck = ((vmin, vmax) != tuple(vlim)) or vlimcheck
		except:
			from ..scriptnotifier.utils import printerr as printerr
			printerr(("Please ensure that the data value limits of all subplots is set to the"
					  " same limits passed to image_grid"), errtype='warning')
		if vlimcheck:
			raise ValueError("The single colorbar is being set to a limit of `%s`; however, a plot"
							 " was found to have a limit of `%s`" % (vlim, (vmin, vmax)))
		cax = grid.cbar_axes[0]
		# Grab the image from the first subplot to pass to the colorbar function
		plot = ax_all[0].get_images()[0]
		vlim, ticklocs, ticklabels = get_contcbarargs(np.array([vlim[0], vlim[1]]),
														 sigfigs=3, vlim=vlim)
		# Plot the colorbar
		cbar = fig.colorbar(plot, cax=cax, ticks=ticklocs)
		# Configure the color bar
		cbar.ax.set_yticklabels(ticklabels, ha='left')
		cbar.ax.tick_params(axis='y', pad=2)
		if cbar_label is not None:
			cbar.set_label(cbar_label, ha='center', va='top', labelpad=2)
	# Export figure
	if output_file or ('pdfpages' in kwargs):
		export_image(output_file, **kwargs)

	return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''fastcomposites.py: Faster compositing with length saving for usage in modeling'''
from __future__ import absolute_import, division, print_function

__author__ = 'pygeostat development team'
__date__ = '2016-12-17'
__version__ = '1.000'

import numpy as np
import pandas as pd
from ..data import DataFile
from numba import jit


def get_composites(composites, datafl, vartypes='continuous', null=-999.0):
    '''Returns a pandas DataFrame with upscaled composites.

    Parameters:
        datafl (DataFile): pygslib DataFile with dh, ifrom, ito and at least 1 variable to upscale
        composites (DataFrame): pandas DataFrame with datafl.dh, datafl.ifrom, datafl.ito
        corresponding to the composite locations

    Keyword Args:
        vartypes (str OR dict): 'continuous', 'categorical' or a dictionary of variables like:
        {'Cu':'continuous','Facies','categorial'}
        null: the null value (values less than or equal to this will not be used)

    Returns:
        upscaled (DataFrame): Pandas DataFrame with values of upscaled variable

    .. codeauthor: pygeostat development team 2014-04-03'''

    # Get a list of drill holes for composites
    dhids = sorted(list(set(composites[datafl.dh])))

    # Check that all drillholes in the composite list are in the datafl
    datafl_dhids = list(set(datafl.data[datafl.dh]))
    for dhid in dhids:
        assert(dhid in datafl_dhids), 'ERROR: {} not in datafl'.format(dhid)

    # Determine the number of composites
    ncomposites = len(composites[datafl.dh])

    # Create a dictionary of variable upscaling methods if not supplied
    if isinstance(vartypes, str):
        vartypedict = {}
        for var in datafl.data.columns:
            if var not in [datafl.dh, datafl.ifrom, datafl.ito]:
                vartypedict[var] = vartypes.lower()
    else:
        vartypedict = vartypes

    # Determine number of variables
    nvar = len(vartypedict)

    # Convert categorical to numbers
    cat_to_numeric_dict = {}
    for var, vartype in vartypedict.items():
        if vartype.startswith('cat'):
            catlist = [cat for cat in list(set(datafl.data[var])) if cat != null]
            catdict = dict((cat, i) for cat, i in zip(catlist, range(1,len(catlist)+1)))
            catdict[null] = null
            cat_to_numeric_dict[var] = catdict
            datafl.data[var] = datafl.data[var].apply(catdict.get)


    # Group data by drill hole ID
    grouped_composites = composites.groupby([datafl.dh], squeeze=True, as_index=False)
    grouped_data = datafl.data.groupby([datafl.dh], squeeze=True, as_index=False)

    upscaled = composites.copy()

    for var, vartype in vartypedict.items():
        comp_arrs = []
        for dhid, compdf in grouped_composites:
            try:
                datadf = grouped_data.get_group(dhid)
            except KeyError:
                raise Exception('No sample data for {}'.format(dhid))

            # Composited array needs to be N x 4 (from, to, variable value, comp length)
            comp_arr = np.zeros((len(compdf), 4))
            comp_arr[:,:-2] = compdf[[datafl.ifrom, datafl.ito]].values

            # Data array only contains from, to, variable value
            data_arr = datadf[[datafl.ifrom, datafl.ito, var]].values

            # Composite and get the value and length
            comp_arr[:,2:] = np.apply_along_axis(comp_value_and_length,
                                      1,
                                      comp_arr,
                                      data_arr,
                                      vartype,
                                      null)
            comp_arrs.append(comp_arr)
        comp_arr = np.concatenate(comp_arrs)
        upscaled[var] = comp_arr[:,2]
        upscaled[var+'_length'] = comp_arr[:,3]

    # Map categorical variables back
    for var, vartype in vartypedict.items():
        if vartype.startswith('cat'):
            revdict = {v:k for k,v in cat_to_numeric_dict[var].items()}
            datafl.data[var] = datafl.data[var].apply(revdict.get)
            upscaled[var] = upscaled[var].apply(revdict.get)

    return(upscaled)

@jit(nopython=True)
def row_length_in_interval(datarow, compfrom, compto, null):
    'Finds the length in the interval along the drill hole - tda'
    # Is the variable value valid?
    if datarow[2] == null:
        return 0.0
    # Case 1 - entire sample interval in composite
    if (datarow[0] >= compfrom) and (datarow[1] <= compto):
        return datarow[1] - datarow[0]
    # Case 2 - partially bottom of interval
    elif ((datarow[0] >= compfrom) and (datarow[0] < compto) and
          (datarow[1] > compto)):

        return compto - datarow[0]
    # Case 3 - partially in top of interval
    elif ((datarow[0] < compfrom) and (datarow[1] > compfrom) and
          (datarow[1] <= compto)):

        return datarow[1] - compfrom
    # Case 4 - composite bracketed by sample interval
    elif (datarow[0] <= compfrom) and (datarow[1] >= compto):
        return compto - compfrom
    # Case 5 - no length in interval
    else:
        return 0.0

@jit
def comp_value_and_length(comprow, data_arr, vartype, null):
    '''Return the composited value and length'''
    lengths = np.apply_along_axis(row_length_in_interval,
                                  1,
                                  data_arr,
                                  comprow[0],
                                  comprow[1],
                                  null)
    total_length = np.sum(lengths)
    if np.sum(lengths) <= 0.0:
        return((null, total_length))
    if vartype.startswith('con'):
        return((np.dot(data_arr[:,2], lengths)/total_length), total_length)
    else:
        unique_cats = list(set(data_arr[lengths > 0.0, 2]))
        if len(unique_cats) == 1:
            return((unique_cats[0], total_length))
        else:
            unique_cat_wts = np.zeros(len(unique_cats))
            # Note: list comprehension is not used to permit usage of JIT
            for i, cat in enumerate(unique_cats):
                unique_cat_wts[i] = np.sum(lengths[data_arr[:,2] == cat])
            unique_cat_wts = unique_cat_wts.tolist()
        return((unique_cats[unique_cat_wts.index(max(unique_cat_wts))], total_length))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Function that emulates the GSLIB program plotem when used with slice_plot"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import ImageGrid

from . set_style import set_plot_style
from .. pygeostat_parameters import Parameters


@set_plot_style
def grid_slice_plot(data, griddef = None, super_xlabel=True, super_ylabel=True, super_title=None, ncol=None, nrow=None,
                    start_slice=None, end_slice=None, figsize=None, n_slice=None, slice_title=True, unit=None,
                    plot_style=None, custom_style=None, output_file=None, out_kws=None, cbar_label=None,
                    axpad=0.15, cbar_cats=None, ntickbins=None, axfuncs=None, label_mode='L', **kwargs):
    """
    grid_slice_plot can be used to automatically generate set of slices
    through a 3D gridded realization. Given some target number of rows, columns, orientation and
    slice ranges through the realization, this function will automatically generate the slice_plot
    slices and arrange them according to the specified dimensions. It is possible to pass keyword
    arguments for slice_plot to this function in order to specify the format of the slice_plots. So,
    adding data locations, different colormaps, and other slice_plot formatting is permitted for
    all subplots by passing those slice_plot arguments to this function. See
    :func:`gs.slice_plot()<pygeostat.plotting.slice_plot>` for the full list of permissable kwargs.

    Updates April 2016 - use a ImageGrid subplots to get the job done

    Parameters:
        data (array, dataframe) : array of data, passed directly to slice_plot()
        griddef (pygeostat griddef) : pygeostat grid definitions, passed directly to slice_plot()
        super_xlabel (str) :  super x axis label
        super_ylabel (str) :  super y axis label
        super_title (str) : super title for the subplots
        ncol (int) : the number of columns considered for the subplots (may change)
        nrow (int) : the number of rows considered for the subplots (may change)
        start_slice (int) : the starting slice to be plotted
        end_slice (int) : the end slice to be plotted
        figsize (tuple) : size of the figure to be created
        n_slice (int) : the number of desired slices
        slice_title (bool) : either plot the orientation and slice no on top of each slice, or dont!
        unit (str): Unit to place inside the axis label parentheses
        plot_style (str) : Use a predefined set of matplotlib plotting parameters as specified by
            :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
            to turn it off
        custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
        output_file (str): Output figure file name and location
        out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
            :func:`gs.exportimg() <pygeostat.plotting.exportimg.exportimg>`
        cbar_label (str): colorbar title
        axpad (float): figure points padding applied to the axes, vertical padding is further
            modified to account for the slice titles, if required
        ntickbins (int or tuple): The number of tick bins for both axes, or the (x, y) respectively
        axfuncs (function or list of functions): External function(s) that takes `ax` `slice_number` and
            `orient` as keyword arguments, does not return anything
        label_mode (str): default `'L'`, or `'all'`, passed to the ImageGrid() constructor
        **kwargs  : **NOTE** the arguments here are either valid slice_plot (including all keyword)
            dictionary arguments, or valid imshow and valid imshow keyword arguments. If errors are
            thrown from invalid arguments it is likely that something that shouldnt have been
            passed to imshow was passed. Check and double check those \**kwargs!

    Returns:
        fig (plt.figure) : figure handle


    **Examples:**

    A simple call: generates a set of slices through the model

    .. plot::

        import pygeostat as gs
        # load some data
        data = gs.ExampleData('3d_grid',griddef = gs.GridDef([40,1,2,40,1,2,40,0.5,1]))
        # plot the grid slices
        _ = gs.grid_slice_plot(data)

    |

    Possible to specify the orientation and the number of slices:

    .. plot::

        import pygeostat as gs
        # load some data
        data = gs.ExampleData('3d_grid',griddef = gs.GridDef([40,1,2,40,1,2,40,0.5,1]))
        # plot the grid slices
        _ = gs.grid_slice_plot(data, orient='xz', n_slice=5)
        

    |

    Can specify the number of rows or columns required for the slices:

    .. plot::

        import pygeostat as gs
        # load some data
        data = gs.ExampleData('3d_grid',griddef = gs.GridDef([40,1,2,40,1,2,40,0.5,1]))
        # plot the grid slices
        _ = gs.grid_slice_plot(data, orient='xz', n_slice=6, ncol=2, nrow=3)

    |

    Also able to specify slice_plot kwargs using this function, so we can apply consistent
        custom formatting to all of the subplots:

    .. plot::
    
        import pygeostat as gs
        # load some data
        data = gs.ExampleData('3d_grid',griddef = gs.GridDef([40,1,2,40,1,2,40,0.5,1]))
        # plot the grid slices
        _ = gs.grid_slice_plot(data, nrow=2, ncol=5, start_slice=10, end_slice=25, n_slice=10, cmap='hot', vlim=(-3,3))

    """
    from . slice_plot import slice_plot
    from .utils import (addxticks, get_contcbarargs, get_supaxislocs)
    from ..datautils.utils import slice_grid
    from ..utility.logging import printerr
    from .export_image import export_image

    from ..data.data import DataFile

    if axfuncs is not None and not isinstance(axfuncs, list):
        axfuncs = [axfuncs]
    if out_kws is None:
        out_kws = dict()

    if griddef is None:
        if isinstance(data, DataFile):
            griddef = data.griddef
        else:
            raise ValueError ('griddef must be provided')
    
    if isinstance(data, DataFile):
        data = data.data

    if 'orient' in kwargs:
        ori = kwargs.get('orient')
        del kwargs['orient']
    else:
        ori = 'xy'
    if slice_title and 'title' in kwargs:
        del kwargs['title']
    if 'slice_number' in kwargs:
        del kwargs['slice_number']
    if 'ax' in kwargs:
        del kwargs['ax']

    if 'catdata' in kwargs:
        catdata = kwargs.get('catdata')
    else:
        if len(np.unique(data)) <= 12:
            catdata = True
        else:
            catdata = False

    if start_slice is None or start_slice < 0:
        start_slice = 0

    if hasattr(data, 'values'):
        data = np.array(data)

    di = ori[1]
    if end_slice is None or end_slice > getattr(griddef, 'n' + di):
        if ori is 'xy':
            end_slice = griddef.nz
        elif ori is 'xz':
            end_slice = griddef.ny
        elif ori is 'yz':
            end_slice = griddef.nx

    if end_slice <= 0:
        print('Bad end_slice parameter, defaulting to 1')
        end_slice = 1

    if n_slice is None:
        n_slice = 4

    if nrow is None:
        addrows = True
        nrow = 2
    else:
        addrows = False
    if ncol is None:
        addcols = True
        ncol = 2
    else:
        addcols = False

    # Try to automatically expand the rows,columns if we dont have enough for the number of slices
    alternate = 1
    if not addrows and not addcols and nrow * ncol < n_slice:

        printerr('Please ensure there are enough rows and columns explicitly defined for '
                 'the target number of plots! (n_slice = %i, nrow*ncol = %i)' %
                 (n_slice, nrow * ncol))
        raise Exception()
    while ncol * nrow < n_slice:
        if alternate == 1:
            if addcols:
                ncol += 1
            alternate = 0
        else:
            if addrows:
                nrow += 1
            alternate = 1

    # Deal with some of the kwarg inputs (add any other defaults that are needed?)
    if 'sigfigs' in kwargs:
        sigfigs = kwargs.get('sigfigs')
    else:
        sigfigs = 3

    # set the vlim to the min max of the entire dataset if none is provided
    if 'vlim' in kwargs:
        vlim = kwargs.get('vlim')
        del kwargs['vlim']
    else:
        vlim = (np.nanmin(data), np.nanmax(data))

    # Turn off slice_plot axis labels if we want super axis labels:
    if super_xlabel:
        kwargs['xlabel'] = False
    if super_ylabel:
        kwargs['ylabel'] = False

    # Create the subplots on the new ImageGrid
    if figsize is None:
        figsize = mpl.rcParams.get("figure.figsize")  # but also needed for later
    fig = plt.figure(figsize=figsize)  # this should just pull from rcParams?
    if slice_title:
        if isinstance(mpl.rcParams['axes.titlesize'], str):
            scalar = 1.1
        else:  # assume this operation will work..
            scalar = mpl.rcParams['axes.titlesize'] / 8
        vaxpad = (axpad * scalar) + 0.1
    else:
        vaxpad = axpad
    imgrid = ImageGrid(fig, 111, nrows_ncols=(nrow, ncol), axes_pad=(axpad, vaxpad),
                       label_mode=label_mode, direction='row', cbar_mode='single', cbar_size=0.15)

    axind = 0
    # ensure a suitable number of slices
    steps = max(int(np.floor((end_slice - start_slice) / n_slice)), 1)
    slrange = np.arange(start_slice, end_slice, steps)[0:n_slice]

    num_slices = 0
    num_errors = 0
    for sl in slrange:
        num_slices += 1
        if slice_title:
            ptitle = ori + ' slice at: ' + str(round(griddef.get_slice_coordinate(ori, sl), 1))
        else:
            ptitle = ""
        if axind < nrow * ncol:
            try:
                ax, plot = slice_plot(data, griddef, ax=imgrid[axind], slice_number=sl, orient=ori,
                                      vlim=vlim, title=ptitle, cbar=False,return_plot=True, plot_style=False,
                                       output_file=None, out_kws=None, **kwargs)
                if ntickbins is not None:
                    if isinstance(ntickbins, tuple):
                        ax.locator_params(axis='x', nbins=ntickbins[0])
                        ax.locator_params(axis='y', nbins=ntickbins[1])
                    else:
                        ax.locator_params(nbins=ntickbins)
                if axfuncs:
                    for func in axfuncs:
                        func(slice_number=sl, ax=imgrid[axind], orient=ori)
                axind += 1
            except ValueError as error:  # hopefully this error was raised by an empty slice..
                num_errors += 1
                err = error
                pass

    if num_errors == num_slices:
        raise ValueError(err)

    # remove unused axes from the figure
    [x.set_axis_off() for x in imgrid[axind:]]

    # Add xticklabels to the last subplot in a column when it is not in the last row
    addxticks(fig, nrow, ncol, axind, griddef.extents(orient=ori[0]))

    # the vlim is either given or set to the max range for the entire array of data..
    # could get more clever about this by preanalyzing the data slices and determining the
    # correct range from all the slices...
    if catdata:  # this section is straight from the slice_plot section for categorical variables
        ticklabels = np.unique(data[~np.isnan(data)])
        if isinstance(cbar_cats, dict):
            ticklabels = [cbar_cats[cat] for cat in ticklabels]
        else:
            ticklabels = [int(x) for x in ticklabels]
        ncat = len(ticklabels)
        vlim = (0, ncat)
        ticklocs = np.arange(ncat) + 0.5
    else:
        vlims, ticklocs, ticklabels = get_contcbarargs(slice_grid(data, griddef, ori, sl),
                                                          sigfigs, vlim)
    # get the min and max colorbar arguments
    cbar = fig.colorbar(plot, imgrid.cbar_axes[0], ticks=ticklocs)
    cbar.ax.set_yticklabels(ticklabels, ha='left')
    cbar.ax.tick_params(axis='y', pad=2)
    if cbar_label is not None:
        cbar.set_label(cbar_label, ha='center', va='top', labelpad=2)

    # Get the figure coordates for the superaxis titles and the figure title
    if super_xlabel or super_ylabel or super_title:
        xmin, xmid, ymin, ymid, ymax = get_supaxislocs(fig, nrow, ncol, figsize, 0.01)

    if unit is None:
        unit = Parameters['plotting.unit']
    if unit is None or unit == '':
        unit = ''
    else:
        unit = ' ({})'.format(unit)
    if super_xlabel:
        if not isinstance(super_xlabel, str):
            if ori is 'xy':
                super_xlabel = Parameters['plotting.xname'] + unit
            elif ori is 'xz':
                super_xlabel = Parameters['plotting.xname'] + unit
            elif ori is 'yz':
                super_xlabel = Parameters['plotting.yname'] + unit
        plt.draw()  
        fig.text(xmid, ymin, super_xlabel, va='top', ha='center')
    if super_ylabel:
        if not isinstance(super_ylabel, str):
            if ori is 'xy':
                super_ylabel = Parameters['plotting.yname'] + unit
            elif ori is 'xz':
                super_ylabel = Parameters['plotting.zname'] + unit
            elif ori is 'yz':
                super_ylabel = Parameters['plotting.zname'] + unit
        plt.draw()
        fig.text(xmin, ymid, super_ylabel, ha='right', rotation=90, va='center')
    if super_title:
        plt.draw()
        fig.text(xmid, ymax, super_title, va='bottom', ha='center')

    # Export figure
    if output_file or ('pdfpages' in out_kws):
        export_image(output_file=output_file, **out_kws)

    return fig



'''
Python interface for geostatistics
==================================

pygeostat is a Python module for geostatistics including interfacing
with GSLIB (Geostatistical Software Library, Copyright 1992-2009
Clayton Deutsch and Andre Journel) and CCG (Centre for Computational
Geostatistics, Copyright 1998-2014 individual authors) software.
It aims to provide a very efficient method to run, script and
interact with GSLIB and CCG software using Python.
'''
from __future__ import absolute_import, division, print_function

__version__ = '1.0.5'

# parameter configurations
from . pygeostat_parameters import Parameters, PlotStyle

## Load necessary modules
from .data import *
from .programs import *
from .plotting import *
from .datautils import *
from .transformations import *
from .multivariate import *
from .statistics import *
from .utility import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Kernel Density Estimation (KDE) calculation module """
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
from scipy.stats import gaussian_kde
from statsmodels.nonparametric.kde import KDEUnivariate
from statsmodels.nonparametric.kernel_density import KDEMultivariate
import numpy as np

def kde_scipy(x, x_grid, bandwidth=0.2, **kwargs):

    """Kernel Density Estimation based on different packages and different kernels
    Note that scipy weights its bandwidth by the covariance of the
    input data.  To make the results comparable to the other methods,
    we divide the bandwidth by the sample standard deviation here.
    """

    if bandwidth > 0:
        kde = gaussian_kde(x, bw_method=bandwidth / x.std(ddof=1), **kwargs)
    else:
        kde = gaussian_kde(x, **kwargs)

    return kde.evaluate(x_grid)


def kde_statsmodels_u(x, x_grid, bandwidth=0.2, **kwargs):
    """Univariate Kernel Density Estimation with Statsmodels"""
    kde = KDEUnivariate(x)
    kde.fit(bw=bandwidth, **kwargs)
    return kde.evaluate(x_grid)


def kde_statsmodels_m(x, x_grid, bandwidth=0.2, **kwargs):
    """Multivariate Kernel Density Estimation with Statsmodels"""
    kde = KDEMultivariate(x, bw=bandwidth * np.ones_like(x),
                          var_type='c', **kwargs)
    return kde.pdf(x_grid)


def kde_sklearn(x, x_grid, bandwidth=0.2, **kwargs):
    """Kernel Density Estimation with Scikit-learn"""
    try:
        from sklearn.neighbors import KernelDensity
    except ImportError:
        raise ImportError("The plotting library Scikit-learn was not found, please install it to to enable"
                            " kde_sklern. It can be installed via Anaconda by executing the following"
                            " command in either cygwin or windows command prompt:                     "
                            "conda install Scikit-learn")
    kde_skl = KernelDensity(bandwidth=bandwidth, **kwargs)
    kde_skl.fit(x[:, np.newaxis])
    # score_samples() returns the log-likelihood of the samples
    log_pdf = kde_skl.score_samples(x_grid[:, np.newaxis])
    return np.exp(log_pdf)

# The Scipy KDE implementation contains only the common Gaussian Kernel. Statsmodels contains seven kernels,
# while Scikit-learn contains six kernels, each of which can be used with one of about a dozen distance metrics,
# resulting in a very flexible range of effective kernel shapes.

kde_funcs = [kde_statsmodels_u, kde_statsmodels_m, kde_scipy, kde_sklearn]
kde_funcnames = ['Statsmodels-U', 'Statsmodels-M', 'Scipy', 'Scikit-learn']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''A collection of utility functions for data prosessing and manipulation'''
from __future__ import absolute_import, division, print_function

import os
import math
import copy
import numpy as np
import pandas as pd
from ..data import GridDef as GridDef
from ..data import DataFile as DataFile

def round_sigfig(value, sigfigs):
    """
    Round a float or integer to a specified number of significant figures. Also handles effectively
    zero, infinity, and negative infinity values.\n
    From: http://stackoverflow.com/questions/3410976/

    Parameters:
        value (int or float): Value that requires rounding
        sigfigs (int): Number of significant figures to round the value to

    Returns:
        new_value (int or float): Rounded value

    Example:
        >>> gs.round_sigfig(-0.00032161, 3)
        >>> -0.00322

    .. codeauthor:: pygeostat development team 2015-10-13
    """
    if value == 0 or math.isnan(value):
        return value
    else:
        new_value = round(value, -int(math.floor(math.log10(abs(value)))) + (sigfigs - 1))
    if abs(new_value) < 1e-15 and abs(new_value) >= 0:
        new_value = 0.00
    elif new_value > 1e+15:
        new_value = float("inf")
    elif new_value < -1e+15:
        new_value = float("-inf")

    return new_value


def fileheader(datafl, mute=False):
    """
    Read a GSLIB file from python and return the header information. Useful for large files.

    .. codeauthor:: pygeostat development team 2016-02-15
    """
    # Get the header and print out if required
    nvar = 999
    variables = []
    with open(datafl) as file:
        for i, line in enumerate(file):
            line = line.replace('\n', '')
            if i == 0:
                if not mute:
                    print(line)
                header = line
            elif i == 1:
                if not mute:
                    print(line)
                line = line.split()
                nvar = int(line[0])
            elif i <= nvar + 1:
                if not mute:
                    print(line)
                variables.append(line)
            else:
                break
            i += 1
    # Return the data if required
    if mute:
        return header, nvar, variables


def corrmatstr(corrmat, fmt):
    """
    Converts a correlation matrix that is currently a numpy matrix or a pandas dataframe, into a
    space delimited string. Correlation matrix strings are required in the parameter files of CCG
    programs such as USGSIM and supersec.

    Currently, this function is hard coded to return two formats as specified by the ``fmt``
    argument, one for ``'usgsim'`` and one for ``'supersec'``. ``'usgsim'`` returns the full
    correlation matrix while ``'supersec'`` returns only the upper triangle of the matrix, without
    the diagonal values.

    Parameters:
        corrmat: Correlation matrix as either a pandas dataframe (pd.DataFrame) or numpy matrix
            (np.ndarray).
        fmt (str): Indicate which format to return. Accepts only one of ``['usgsim', 'supersec']``

    Returns:
        corrstr (str): Correlation matrix as a space delimited string.

    .. codeauthor:: pygeostat development team 2016-03-15
    """
    # Sanity Check
    if fmt not in ['supersec', 'usgsim']:
        return ValueError("One of 'supersec' or 'usgsim' must be passed")
    # Convert the correlation matrix to a string
    if not isinstance(corrmat, pd.DataFrame):
        corrmat = pd.DataFrame(data=corrmat)
    if fmt is 'supersec':
        if isinstance(corrmat, pd.DataFrame):
            corrmat = corrmat.as_matrix()
        ind = np.triu_indices(len(corrmat), 1)
        i = 0
        corrstr = ""
        for j in range(len(corrmat) - 1, 0, -1):
            for k in range(j):
                corrstr = corrstr + "%s " % corrmat[ind][i]
                i += 1
            corrstr = corrstr + "\n"
    elif (fmt is 'usgsim') or fmt:
        corrstr = ""
        for i in range(0, len(corrmat)):
            string = (' '.join(map(str, corrmat.ix[i, :])))
            corrstr = corrstr + string + '\n'

    return corrstr


def slice_grid(data, griddef, orient, slice_number, slice_thickness=None, tmin=None):
    """
    Slice a 3-D grid.

    Parameters:
        data: 1-D array or a tidy long-form dataframe with a single column containing the variable
            in question and each row is an observation
        griddef (GridDef): A pygeostat GridDef class created using
            :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`
        orient (str): Orientation to slice data. ``'xy'``, ``'xz'``, ``'yz'`` are the only accepted
            values
        slice_number (int): Grid cell location along the axis not plotted to take the slice of data to
            plot
        slice_thickness (int): Number of slices around the selected slice_number to average the attributes
        tmin (float): A minimum threshold value to filter/key out the data

    Returns:
        view (np.ndarray): 1-D array of the sliced data

    .. codeauthor:: pygeostat development team 2014-04-19
    """
    if not isinstance(slice_number, int):
        slice_number = int(slice_number)
    if data.shape != (griddef.nz, griddef.ny, griddef.nx):
        data = data.reshape((griddef.nz, griddef.ny, griddef.nx))
    if orient == 'xy':
        view = data[slice_number, :, :]
    elif orient == 'xz':
        view = data[:, slice_number, :]
    elif orient == 'yz':
        view = data[:, :, slice_number]
    else:
        raise ValueError("Not a valid orientation! {}".format(orient))
    # new functionality to get the average of a set of slices, honoring null values if present
    if slice_thickness is not None:
        slicerange = np.arange(slice_number - slice_thickness, slice_number + slice_thickness)
        nslice = len(slicerange)
        slices = np.zeros((nslice, *view.shape), dtype=np.float)
        for isl, slno in enumerate(slicerange):
            if orient == 'xy':
                slices[isl, :, :] = data[slno, :, :]
            elif orient == 'xz':
                slices[isl, :, :] = data[:, slno, :]
            elif orient == 'yz':
                slices[isl, :, :] = data[:, :, slno]
        if tmin is None:
            view = np.mean(slices, axis=0)
        else:
            slices = np.ma.masked_array(slices, mask=(slices <= tmin))
            view = np.ma.mean(slices, axis=0)
    return view


def slicescatter(data, orient, slice_number, slicetol, griddef=None, x=None, y=None, z=None):
    """
    Slice scattered data based on a GSLIB style grid definition.

    Parameters:
        data (pd.DataFrame or gs.DataFile): Dataframe where each column is a variable and each row
            is an observation. Must contain the coordinate columns required depending on the value
            of ``orient``. If a :class:`gs.DataFile <pygeostat.data.data.DataFile>` class is
            passed, its attribute ``griddef``, ``x``, ``y``, and ``z`` will be extracted.
        var (str): Column header of variable under investigation
        orient (str): Orientation to slice data. ``'xy'``, ``'xz'``, ``'yz'`` are the only accepted
            values
        slice_number (int): Grid cell location along the axis not plotted to take the slice of data to
            plot
        slicetol (float): Slice tolerance to plot point data (i.e. plot +/- ``slicetol`` from the
            center of the slice). Any negative value plots all data. Default is to plot all data.
        griddef (gs.GridDef): A pygeostat GridDef class created using
            :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Required if the
            attribute cannot be retrieved from ``data`` if it is a
            :class:`gs.DataFile <pygeostat.data.data.DataFile>` class.
        x (str): Column header of x-coordinate. Required if the attribute cannot be retrieved from
            ``data`` if it is a :class:`gs.DataFile <pygeostat.data.data.DataFile>` class.
        y (str): Column header of x-coordinate. Required if the attribute cannot be retrieved from
            ``data`` if it is a :class:`gs.DataFile <pygeostat.data.data.DataFile>` class.
        z (str): Column header of x-coordinate. Required if the attribute cannot be retrieved from
            ``data`` if it is a :class:`gs.DataFile <pygeostat.data.data.DataFile>` class.

    Returns:
        pointview (pd.DataFrane): pd.DataFrame of the sliced data
    .. codeauthor:: pygeostat development team - 2016-04-11
    """
    # Handle default parameters
    if isinstance(data, DataFile):
        if not griddef:
            if data.griddef:
                griddef = data.griddef
        if not x:
            x = data.x
        if not y:
            y = data.y
        if not z:
            z = data.z
        data = data.data
    if not x:
        if 'x' in data.columns:
            x = 'x'
    if not y:
        if 'y' in data.columns:
            y = 'y'
    if not z:
        if 'z' in data.columns:
            z = 'z'
    # Sanity checks
    if not any([isinstance(data, pd.DataFrame), isinstance(data, DataFile)]):
        raise ValueError("The parameter data must be either a gs.DataFile or a pd.DataFrame")
    if orient is 'xy':
            if any([x is None, y is None]):
                raise ValueError("`orient` is set to `'xy'` yet the column ID for `x` and `y` has"
                                 " not been set and could not be automatically retrieved.")
    elif orient is 'xz':
        if any([x is None, z is None]):
            raise ValueError("`orient` is set to `'xz'` yet the column ID for `x` and `z` has"
                             " not been set and could not be automatically retrieved.")
    elif orient is 'yz':
        if any([y is None, z is None]):
            raise ValueError("`orient` is set to `'yz'` yet the column ID for `y` and `z` has"
                             " not been set and could not be automatically retrieved.")
    else:
        raise ValueError("An `orient` setting of %s is not permissible" % orient)
    if not isinstance(griddef, GridDef):
        raise ValueError("A valid gs.GridDef was not passed")
    # Slice the data
    if orient is 'xy':
        if griddef.nz > 1:
            pointview = data[data[z] >= (griddef.get_slice_coordinate('xy', slice_number) - slicetol *
                                         griddef.zsiz)]
            pointview = pointview[pointview[z] <=
                                  (griddef.get_slice_coordinate('xy', slice_number) + slicetol * griddef.zsiz)]
    elif orient is 'xz':
        if griddef.ny > 1:
            pointview = data[data[y] >= (griddef.get_slice_coordinate('xz', slice_number) - slicetol *
                                         griddef.ysiz)]
            pointview = pointview[pointview[y] <=
                                  (griddef.get_slice_coordinate('xz', slice_number) + slicetol * griddef.ysiz)]
    elif orient is 'yz':
        if griddef.nx > 1:
            pointview = data[data[x] >= (griddef.get_slice_coordinate('yz', slice_number) - slicetol *
                                         griddef.xsiz)]
            pointview = pointview[pointview[x] <=
                                  (griddef.get_slice_coordinate('yz', slice_number) + slicetol * griddef.xsiz)]

    return pointview


def fixpath(path):
    """
    Convert a file path to an absolute path if required and make sure there are only forward
    slashes.

    If copying the path directly from windows explorer or something that will produce a path like
    that, make sure to indicate to python that the string is raw. This is done by placing a ``r``
    in front of the string. For example:

    >>> string = r"A string with backslashes \\ \\ \\ \\"

    Example:
        Make sure to place an ``r`` in front of the string so funny things don't happen. A simple
        call:

        >>> gs.fixpath(r"D:\\Data\\data.dat")

    .. codeauthor:: pygeostat development team - 2016-02-07
    """
    # Convert any relative paths to absolute paths if needed
    if not os.path.isabs(path):
        path = os.path.abspath(path)
    # Fix slashes if required
    path = path.replace('\\\\\\', '\\')
    path = path.replace('\\\\', '\\')
    path = path.replace('\\', '/')

    return path


def is_numeric(s):
    """Returns true if a value can be converted to a floating point number"""
    try:
        float(s)
        return True
    except ValueError:
        return False
    except TypeError:
        print('ERROR: Must have null value in working dictionary')
        return False


def ensure_dir(f):
    """Function to make sure that directory(s) exists and if not, create it"""
    import os, copy
    if isinstance(f, str):
        dirs = [f]
    else:
        dirs = copy.copy(f)
    for f in dirs:
        if not os.path.exists(f):
            os.makedirs(f)

def ensure_path(path):
    """
    Function ensures that all folders in a given path or list of paths are created if they
    do not exist
    """
    if isinstance(path, str):
        path = [path]
    path = [p.replace('\\', '/') for p in path]
    for p in path:
        incremental_path = ''
        dirlist = p.split('/')
        for d in dirlist:
            incremental_path += d + '/'
            if ':' not in d and d != '':
                ensure_dir(incremental_path)

def nearest_eucdist(x, y=None, z=None):
    """
    Calculate the euclidean distance to the nearest sample for each sample.

    Parameters:
        x (np.array): Array of the coordinate in the x direction

    Keyword Arguments:
        y (np.array): Array of the coordinate in the y direction
        z (np.array): Array of the coordinate in the z direction

    Returns:
        dist (np.array): Array of the euclidean distance to the nearest sample for each
        sample

    .. codeauthor:: pygeostat development team - 2016-07-28
    """
    import scipy
    nsamp = len(x)
    coords = []
    if y is not None and z is not None:
        for coord in zip(x, y, z):
            coords.append(coord)
    if y is not None and z is None:
        for coord in zip(x, y):
            coords.append(coord)
    if z is not None and y is None:
        for coord in zip(x, z):
            coords.append(coord)
    distmat = scipy.spatial.distance.cdist(coords, coords, 'euclidean')
    distmat[np.tril_indices(nsamp, k=0, m=nsamp)] = np.nan
    nearest = []
    for i in range(nsamp - 1):
        dist = distmat[i, :]
        nearest.append(np.nanmin(dist))
    nearest = np.array(nearest)

    return nearest



#!/usr/bin/env python
# -*- coding: utf-8 -*

'''filemanagement.py: Contains basic utility functions for creating/removing files and directories'''
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = 'pygeostat development team'
__date__ = '2018'
__version__ = '1.0.0'


def mkdir(dirnames, verbose=False):
    '''
    Make directory(s).

    Parameters:
        dirnames(str or list): directory(s) to create
        verbose(bool): print to screen if an error is encountered,
            including if the dirnames already exists
    '''
    from os import makedirs
    if isinstance(dirnames, str):
        dirnames = [dirnames]
    for dirname in dirnames:
        try:
            makedirs(dirname)
        except:
            if verbose:
                print('{} not created - likely exists already!'.format(dirname))


def rmdir(dirnames, verbose=False):
    '''
    Remove directory(s).

    Parameters:
        dirnames(str or list): directory(s) to remove
        verbose(bool): print to screen if an error is encountered,
            including if the dirnames has already been removed
    '''
    from shutil import rmtree
    if isinstance(dirnames, str):
        dirnames = [dirnames]
    for dirname in dirnames:
        try:
            rmtree(dirname)
        except:
            if verbose:
                print('{} not removed - likely does not exist!'.format(dirname))


def rmfile(filenames, verbose=False):
    '''
    Remove file(s).

    Parameters:
        filenames(str): file to remove
        verbose(bool): print to screen if an error is encountered,
            including if the file has already been removed
    '''
    from os import remove
    if isinstance(filenames, str):
        filenames = [filenames]
    for filename in filenames:
        try:
            remove(filename)
        except:
            if verbose:
                print('{} not removed - likely does not exist!'.format(filename))


def get_executable(source='gslib', access_token=None, clean=True):

    '''
    Gets a collection of executable files from a protected repository using an access token.
    
    Parameters:
        source (str): gslib or CCG as the source of software
        access_token (str): An access token to authorize access to the target private repository
        clean (bool): Option to clean the executable directory prior to upload the files from the target private repository

    **Examples**

    .. code-block:: python
        
        import pygeostat as gs
        gs.get_executable(source='CCG', clean=True)

    |

    '''

    import subprocess
    import os
    import shutil
    import sys
    import getpass

    if access_token is None:
        access_token = getpass.getpass('Please provide a valid access token')

    # check if git is available
    try:
        popenobject_ = subprocess.Popen('git')
        popenobject_.terminate()
    except Exception:
        print ('Unsuccessful! Could not access git. Please make sure git is installed.')

    # Target direcory for executable
    target_dir = os.path.abspath(os.path.join(os.path.dirname( __file__ ), r'..', r'executable'))

    # Clean the directory first before uploading the executable files
    if clean:
        for file_name in os.listdir(target_dir):
            rmfile(os.path.join(target_dir,file_name))

    # Create a temporary directory to get
    temp_dir = 'temp1618'
    mkdir(temp_dir)

    if source.lower() == 'gslib':
        print('The software is available under gslib license agreement (http://www.gslib.com)')
        
        command = 'git clone https://CcgUser:{access_token}@github.com/CcgAlberta/GslibRepository.git {target}'.format(access_token=access_token,target = temp_dir)

        if 'win' in sys.platform:
            source_dir = os.path.join(temp_dir, 'Windows64')
        else:
            source_dir = os.path.join(temp_dir, 'Linux64')
    else:
        print('The software is available under CCG software term of use (http://www.ccgalberta.com/software-terms-of-use)')
        command = 'git clone https://CcgUser:{access_token}@github.com/CcgAlberta/CcgSoftware.git {target}'.format(access_token=access_token,target = temp_dir)

        print('The software is suited for Microsoft platform')
        source_dir = os.path.join(temp_dir, 'CcgFortranExecutable')
        
    try:
        main_process = subprocess.Popen(command, cwd='.',
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                shell=True,
                                bufsize=1,
                                universal_newlines=True)
        rc=main_process.wait()
    except Exception as exception_content:
        try:
            logs = main_process.stdout.read().replace('\r', '')
            main_process.stdout.close()
        except UnboundLocalError:
            logs = str(exception_content)

        logs = logs.replace(access_token, 'retracted')
        
        __remove_temp_dir(temp_dir)
        
        raise Exception(logs) from None

    try:
        for file_name in os.listdir(source_dir):
            
            srcfile = os.path.join(source_dir,file_name)
            
            destfile = os.path.join(target_dir,file_name)
            
            shutil.copy(srcfile,destfile)

    except Exception as exception_content:
        logs = 'Unable to retrive data from the repository. Please make sure correct access token has been provided!' + str(exception_content)
        raise Exception(logs) from None

    finally:
        __remove_temp_dir(temp_dir)


def __remove_temp_dir(temp_dir):
    import os
    import sys
    import stat
    from os import path
    import shutil

    # Remove the temporary directory
    for root, dirs, files in os.walk(temp_dir):  
        for dir in dirs:
            os.chmod(path.join(root, dir), stat.S_IRWXU)
        for file in files:
            os.chmod(path.join(root, file), stat.S_IRWXU)
    shutil.rmtree(temp_dir)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A postsim style utility
"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import pandas as pd
import numpy as np

from ..data import DataFile
from ..data import iotools as iotools

def postsim_multfiles(file_base_or_list, output_name, Nr=None, file_ending=None, fltype=None,
                      output_fltype=None, zero_padding=0, variables=None, var_min=None):
    '''The multiple file postsim function uses recursive statistics for memory management and
    coolness factor. See http://people.revoledu.com/kardi/tutorial/RecursiveStatistic/
    This function will take multiple realizations and post process the results into mean and
    variance for each variable. You can either pass it a list of files to iterate through or a
    filebase name and the number of realizations.

    Parameters:
        file_base_or_list (list) or (str): List of files or path + base name of sequentially named
            files
        output_name (str): ath (or name) of file to write output to.
        Nr (int): Number of realizations. Needed if file base name is passed.
        file_ending (str): file ending (ex. `"out"`). Used if file base name is passed. Period is
            not included.
        fltype (str): Type of data file: either ``csv``, ``gslib``, ``hdf5``, or ``gsb``. Used if
            file base name is passed and `file_ending` is not used.
        output_fltype (str): Type of output data file: either ``csv``, ``gslib``, ``hdf5``, or
            ``gsb``.
        zero_padding (int): Number of zeros to padd number in sequentially named files with. Default
            is 0.
        variables (str): List of variables to process.
        var_min (list) or (float): Minimum trimming limit to use. If one value is passed it will
            apply the trimming limit to all variables. Or a list of trimming limit for each variable
            can be passed.
    '''

    if isinstance(file_base_or_list, list):
        N = 0
        for filename in file_base_or_list:
            N += 1
            dt = DataFile(flname=filename)

            # Create the post sim file in the first call
            if N == 1:
                # setup index length and variables
                blk_count = len(dt.data)
                if variables:
                    if variables not in dt.data.columns.tolist():
                        raise KeyError('Variables passed do not match columns in datafile')
                else:
                    variables = dt.data.columns.tolist()
                columns = list()
                for var in variables:
                    columns.append(var + '_mean')
                    columns.append(var + '_variance')
                columns.append('Nr')
                # create pandas file
                postsim = pd.DataFrame(index=np.arange(blk_count), columns=columns)
                # Set nan values
                if var_min:
                    if isinstance(var_min, list):
                        if len(var_min) != len(variables):
                            raise KeyError('length of var_min list does not equal number of'
                                           ' variables being processed')
                        else:
                            trim_dict = dict(zip(variables, var_min))
                            for var in variables:
                                dt.data.setnan(variables=var, tmin=trim_dict[var])
                    elif isinstance(var_min, (int, float)):
                        dt.data.setnan(variables=variables, tmin=var_min)
                    else:
                        raise KeyError('var_min must be either a list or a number')
                # Initialize the first loop
                postsim.Nr = N
                for var in variables:
                    col_m = var + '_mean'
                    col_v = var + '_variance'
                    postsim[col_m] = dt.data[var]
                    postsim[col_v] = 0
                continue
            # ----Continue with next files----
            # Set nan values
            if var_min:
                if isinstance(var_min, list):
                    if len(var_min) != len(variables):
                        raise KeyError('length of var_min list does not equal number of'
                                       ' variables being processed')
                    else:
                        trim_dict = dict(zip(variables, var_min))
                        for var in variables:
                            dt.data.setnan(variables=var, tmin=trim_dict[var])
                elif isinstance(var_min, (int, float)):
                    dt.data.setnan(variables=variables, tmin=var_min)
                else:
                    raise KeyError('var_min must be either a list or a number')
            # calculate the stats
            postsim.Nr = N
            for var in variables:
                col_m = var + '_mean'
                col_v = var + '_variance'
                # Calculate the arithmetic average
                postsim.left_arg = ((N - 1) / N) * postsim[col_m]
                postsim.right_arg = (1 / N) * dt.data[var]
                postsim[col_m] = postsim.left_arg + postsim.right_arg
                # Calculate the variance
                postsim.left_arg = ((N - 1) / N) * postsim[col_v]
                postsim.temp = dt.data[var]
                postsim.right_arg = postsim.temp - postsim[col_m]
                postsim[col_v] = (postsim.left_arg + (1 / (N - 1)) *
                                  postsim.right_arg * postsim.right_arg)
    elif isinstance(file_base_or_list, str):
        for N in range(1, Nr + 1):
            # Create the post sim file in the first call
            if N == 1:
                if file_ending is None:
                    if fltype.lower() == 'gslib':
                        file_ending = 'out'
                    elif fltype.lower() == 'csv':
                        file_ending = 'csv'
                    elif fltype.lower() == 'gsb':
                        file_ending = 'gsb'
                    elif fltype.lower() == 'h5' or fltype.lower() is 'hdf5':
                        file_ending = 'h5'
                    else:
                        raise KeyError('Either file_ending or fltype is needed')
                filename = '{base}{number}.{ending}'.format(base=file_base_or_list,
                                                            number=str(N).zfill(zero_padding),
                                                            ending=file_ending)
                dt = DataFile(flname=filename)
                # Figure out the index (variable names... and total number of blocks)
                blk_count = len(dt.data)
                if variables:
                    if variables not in dt.data.columns.tolist():
                        raise KeyError('Variables passed do not match columns in datafile')
                else:
                    variables = dt.data.columns.tolist()
                columns = list()
                for var in variables:
                    columns.append(var + '_mean')
                    columns.append(var + '_variance')
                columns.append('Nr')
                columns.append('left_arg')
                columns.append('right_arg')
                columns.append('temp')
                # create pandas dataframe
                postsim = pd.DataFrame(index=np.arange(blk_count), columns=columns)
                # Set nan values
                if var_min:
                    if isinstance(var_min, list):
                        if len(var_min) != len(variables):
                            raise KeyError('length of var_min list does not equal number of'
                                           ' variables being processed')
                        else:
                            trim_dict = dict(zip(variables, var_min))
                            for var in variables:
                                dt.data.setnan(variables=var, tmin=trim_dict[var])
                    elif isinstance(var_min, (int, float)):
                        dt.data.setnan(variables=variables, tmin=var_min)
                    else:
                        raise KeyError('var_min must be either a list or a number')
                # Initialize the first loop
                postsim.Nr = N
                for var in variables:
                    col_m = var + '_mean'
                    col_v = var + '_variance'
                    postsim[col_m] = dt.data[var]
                    postsim[col_v] = 0
                continue
            # ----Continue with next files----
            # open the next file
            filename = '{base}{number}.{ending}'.format(base=file_base_or_list,
                                                        number=str(N).zfill(zero_padding),
                                                        ending=file_ending)
            dt = DataFile(flname=filename)
            # Set nan values
            if var_min:
                if isinstance(var_min, list):
                    if len(var_min) != len(variables):
                        raise KeyError('length of var_min list does not equal number of'
                                       ' variables being processed')
                    else:
                        trim_dict = dict(zip(variables, var_min))
                        for var in variables:
                            dt.data.setnan(variables=var, tmin=trim_dict[var])
                elif isinstance(var_min, (int, float)):
                    dt.data.setnan(variables=variables, tmin=var_min)
                else:
                    raise KeyError('var_min must be either a list or a number')
            # calculate the stats
            postsim.Nr = N
            for var in variables:
                col_m = var + '_mean'
                col_v = var + '_variance'
                # Calculate the arithmetic average
                postsim.left_arg = ((N - 1) / N) * postsim[col_m]
                postsim.right_arg = (1 / N) * dt.data[var]
                postsim[col_m] = postsim.left_arg + postsim.right_arg
                # Calculate the variance
                postsim.left_arg = ((N - 1) / N) * postsim[col_v]
                postsim.temp = dt.data[var]
                postsim.right_arg = postsim.temp - postsim[col_m]
                postsim[col_v] = (postsim.left_arg + (1 / (N - 1)) *
                                  postsim.right_arg * postsim.right_arg)
    else:

        raise TypeError('file_base_or_list must be either a list or a string')
    # Write out the results
    columns.remove('left_arg')
    columns.remove('right_arg')
    columns.remove('temp')
    if output_fltype is None:
        if fltype:
            output_fltype = fltype
        elif file_ending == '.out':
            output_fltype = 'gslib'
        elif file_ending == 'csv' or file_ending == 'gsb' or file_ending == 'h5':
            output_fltype = file_ending
        else:
            raise KeyError('Unable to figure out what output_fltype needs to be. Please pass'
                           ' a value')

    if output_fltype.lower() == 'gslib':
        iotools.write_gslib(postsim, output_name, variables=columns)
    elif output_fltype.lower() == 'csv':
        iotools.write_csv(postsim, output_name, variables=columns)
    elif output_fltype.lower() == 'gsb':
        iotools.write_gsb(postsim, output_name, tvar='Nr', variables=columns)
    elif output_fltype.lower() == 'h5' or fltype.lower() is 'hdf5':
        iotools.write_h5(postsim, output_name, variables=columns)
    else:
        raise NotImplementedError('output_fltype did not match any of the implemented filetypes')



# public
'''
The :mod:`pygeostat.scriptnotifier` module is for notifying the user of script status
or crash reports.
'''
from __future__ import absolute_import, division, print_function

# from .scriptnotifier import ScriptNotifier
from . filemanagement import mkdir, rmdir, rmfile, get_executable
from .logging import printerr, log_progress



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''contour_plot.py: Contains a basic contour plotting routine using matplotlib'''
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import matplotlib.pyplot as plt
import numpy as np

from . set_style import set_plot_style
from .. pygeostat_parameters import Parameters



@set_plot_style
def contour_plot(data, griddef=None, var=None, orient='xy', slice_number=0, ax=None, output_file=None, c='k',
               figsize=None, xlabel=None, ylabel=None, title=None, unit=None, leg_label=None,
               aspect=None, clabel=False, lw=1.0, plot_style=None, custom_style=None, axis_xy=None, grid=None, return_ax=True, return_csi=False):
    """
    Contains a basic contour plotting routine using matplotlib

    Parameters:
        data: A numpy ndarray, pandas DataFrame or pygeostat DataFile, where each column is a
            variable and each row is an observation
        griddef (GridDef): A pygeostat GridDef class, which must be provided if a DataFile is
            not passed as data with a valid internal
            GridDef :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`
        var (str,int): The name of the column within data to plot. If an int is provided, then it
            corresponds with the column number in data. If None, the first column of data
            is used.
        orient (str): Orientation to slice data. ``'xy'``, ``'xz'``, ``'yz'`` are the only accepted
            values
        slice_number (int): Grid cell location along the axis not plotted to take the slice of data to
            plot
        ax (mpl.axis): Matplotlib axis to plot the figure
        output_file (str): Output figure file name and location
        show (bool): ``True`` will use plt.show() at end. Typically don't need this.
        c (str): Matplotlib color
        figsize (tuple): Figure size (width, height)
        xlabel (str): X-axis label
        ylabel (str): Y-axis label
        title (str): title for the plot
        unit (str): Distance unit, taken from Parameters if ``None``
        leg_label (str): Adds a single label to the legend for the contour lines
        aspect (str): Set a permissible aspect ratio of the image to pass to matplotlib.
        clabel (bool): Whether or not to label the contours wth their values
        lw (float): the weight of the contour lines
        plot_style (str): Optional pygeostat plotting style
        custom_style (dict): Custom dictionary for plotting styles
        grid (bool): Plots the major grid lines if True. Based on Parameters['plotting.grid']
            if None.
        axis_xy (bool): converts the axis to GSLIB-style axis visibility (only left and bottom
            visible) if axis_xy is True. Based on Parameters['plotting.axis_xy'] if None.
        return_ax (bool): specify if the plotting axis should be returned
        return_csi (bool): specify if the contour instance should be returned


    Returns:
        csi (ax): Matplotlib ax.contour instance

    **Examples:**

    A basic contour plotting example:

    .. plot::

        import pygeostat as gs
        grid_str = '''120 5.00  10.00 -nx, xmn, xsiz
                    110 1205.00  10.00 -ny, ymn, ysiz
                    1 0.5 1.0  -nz, zmn, zsiz'''
        
        griddef = gs.GridDef(grid_str=grid_str)
        data_fl = gs.ExampleData("grid2d_surf", griddef=griddef)
        gs.contour_plot(data_fl, var="Thickness", clabel=True)

    Contour plot on top of slice plot:

    .. plot::

        import pygeostat as gs
        grid_str = '''120 5.00  10.00 -nx, xmn, xsiz
                    110 1205.00  10.00 -ny, ymn, ysiz
                    1 0.5 1.0  -nz, zmn, zsiz'''
        
        griddef = gs.GridDef(grid_str=grid_str)

        data_fl = gs.ExampleData("grid2d_surf", griddef=griddef)
        ax = gs.slice_plot(data_fl, var="Thickness")
        _ = gs.contour_plot(data_fl, var="Thickness", ax = ax, clabel=True)

    """
    from . utils import format_plot, _spatial_griddata, _spatial_labels, _spatial_aspect
    from .export_image import export_image
    from ..datautils.utils import slice_grid

    # Parse the data, var and griddef input to determine the data and griddef
    data, griddef = _spatial_griddata(data, var, griddef)
    # Slice the data
    if orient in ['xy', 'xz', 'yz']:
        view = slice_grid(data, griddef, orient, slice_number)
    else:
        raise Exception("Error: no orientation set! {}".format(orient))

    a = orient[0]
    b = orient[1]
    xmin = getattr(griddef, a + 'limits')[0]
    xmax = getattr(griddef, a + 'limits')[1]
    ymin = getattr(griddef, b + 'limits')[0]
    ymax = getattr(griddef, b + 'limits')[1]

    if aspect is None:
        aspect = _spatial_aspect([xmin, xmax], [ymin, ymax])

    # Discretize for contour map
    X = np.linspace(xmin, xmax, len(view[0, :]))
    Y = np.linspace(ymin, ymax, len(view[:, 0]))

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=figsize)

    # Create the 'artists' that draw the contours
    csi = ax.contour(X, Y, np.reshape(view, (len(view[:, 0]), len(view[0, :]))), linewidths=lw,
                     colors=c)
    if leg_label is not None:
        ax.plot(np.nan, lw=lw, color=c, label=leg_label)
    if clabel:
        ax.clabel(csi, inline=1, fontsize=9, fmt='%.0f')
    if aspect:
        ax.set_aspect(aspect)
    _spatial_labels(ax, orient, griddef, slice_number, title, xlabel, ylabel, unit)
    if axis_xy is None:
        axis_xy = Parameters['plotting.axis_xy_spatial']
    format_plot(ax, axis_xy=axis_xy, grid=grid)

    # Done Plotting, save figure if required and show it
    if output_file is not None:
        export_image(output_file)

    # return some things
    if return_ax and return_csi:
        return ax, csi
    elif return_ax:
        return ax
    elif return_csi:
        return csi
